/* sx16mod.c
 *  Module    : sx16mod.c
 *  Function  : Linux driver API for Sensoray models 516/616
 *  Authors   : Phillip W. Jim Lamberson, Charlie Liu, Alexander K., Sasha K.,
 *              Dean A.
 *
 *  Note:All new code must pass checkpatch.pl and be in kernel code
 *  style(DA). eg.  80 columns max, K&R style, tab space 8
 *  Copyright (C) Sensoray 2003-2008
 *  Copyright (C) 2003  Sensoray Co., Inc.  This program is free software;
 *  you can redistribute it and/or modify it under the terms of the GNU
 *  General Public License as published by the Free Software Foundation;
 *  either version 2 of the License, or (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *   In addition, as a special exception, Sensoray Co., Inc., gives permission
 *   to link the code of this program with the 'sx16core.o' library
 *   and distribute linked combinations including the two. You must obey the
 *   GNU General Public License in all respects for all of the code used other
 *   than 'sx16core.o'. If you modify this file, you may extend this exception
 *   to your version of the file, but you are not obligated to do so. If you
 *   do not wish to do so, delete this exception statement from your version.
 *
 *   Sensoray makes every effort to support all kernels, but unfortunately
 *   V4L cannot be explicitly supported below 2.6.17.  2.6.9 and below
 *   will never be supported.
 *   The V4L internal API was greatly improved after 2.6.9
 *   so if using V4L on a 2.6.9 or OLDER system,
 *   you must upgrade to a newer kernel.  This includes Linux 2.4.
 *   616 uncompressed V4L capture tested on 2.6.24, 2.6.22, 2.6.18 Ubuntu
 *   Very Important:
 *   Note that the above does NOT apply to the 516 OR if the 616 is compiled
 *   without V4L uncompressed capture support.
 */

#define MODULE_REVISION "2.3.16(January 7, 2021)"
#define MODULE_FILE_BUILD

#include <linux/version.h>
#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 18)
#include <linux/config.h>
#endif

#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 5, 0)
/* 2.4 kernel stuff */
#if defined(CONFIG_MODVERSIONS) && !defined(MODVERSIONS)
#define MODVERSIONS
#endif
#ifdef MODVERSIONS
#include <linux/modversions.h>
#endif
#endif

#include <linux/module.h>
#include <linux/init.h>
#include <linux/kernel.h>
#if LINUX_VERSION_CODE >= KERNEL_VERSION(4, 11, 0)
#include <linux/sched/signal.h>
#endif

#if LINUX_VERSION_CODE >= KERNEL_VERSION(3, 13, 0)
#include <linux/vmalloc.h>
#endif

#if LINUX_VERSION_CODE > KERNEL_VERSION(2, 5, 0)
#include <linux/wait.h>
#endif
#include <linux/fs.h>
#include <linux/proc_fs.h>
#if LINUX_VERSION_CODE >= KERNEL_VERSION(3, 10, 0)
#include <linux/seq_file.h>
#endif
#include <linux/pci.h>
#include <linux/poll.h>
#include <linux/delay.h>
#include <linux/sched.h>
#include <linux/interrupt.h>
#include <linux/pagemap.h>
#if LINUX_VERSION_CODE > KERNEL_VERSION(2, 6, 17)
#include <linux/uaccess.h>
#include <linux/io.h>
#endif
#include <linux/irq.h>
#include <linux/spinlock.h>

#define KERNEL_VERSION_NOINODE KERNEL_VERSION(2, 6, 29)

#ifdef SX16_V4L_BUILD
#include <linux/videodev2.h>

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 24)
#include <media/videobuf-dma-sg.h>
#else
#include <media/video-buf.h>
#endif

#include <media/v4l2-dev.h>

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 28)
#include <media/v4l2-ioctl.h>
#include <media/v4l2-device.h>
#endif

#if LINUX_VERSION_CODE >= KERNEL_VERSION(4, 8, 0)
#include <linux/compat.h>
#include <linux/time.h>
#endif






/*
 *  warning message for support.
 *  Sensoray tests on the latest Ubuntu and Ubuntu LTS
 *  releases only. You may have to debug V4L APIs on other kernels
 */

#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 18)
#error "This older version of the Linux kernel is NOT supported"
#error "for V4L uncompressed 616 capture"
#error "Please upgrade your kernel if you are using the 616 for"
#error "uncompressed capture"
#error "The V4L API has important bug fixes in later kernels"
#endif

#if LINUX_VERSION_CODE >= KERNEL_VERSION(5, 11, 0)
#error "Sensoray has not tested the 616 V4L API on this kernel yet"
#error "You may try uncommenting this line and recompiling"
#error "but this kernel is not officially supported yet"
#endif

#define KERNEL_VERSION_V4L2DEV KERNEL_VERSION(2, 6, 29)

#if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,25)
#define __devinit 
#endif


/* convert for 2.6.24 and lower */
#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 25)
#define VIDEOBUF_DONE       STATE_DONE
#define VIDEOBUF_ACTIVE     STATE_ACTIVE
#define VIDEOBUF_QUEUED     STATE_QUEUED
#define VIDEOBUF_ERROR      STATE_ERROR
#define VIDEOBUF_NEEDS_INIT STATE_NEEDS_INIT
#define VIDEOBUF_PREPARED   STATE_PREPARED
#endif

#endif



	
	
/* Threading stuff... */
#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 5, 0)
/* 2.4 kernel */
#include "kthread.h"
#else
/* 2.6 kernel */
#include <linux/moduleparam.h>
#include <linux/kthread.h>
#endif

#define KERNEL_VERSION_NOINODE KERNEL_VERSION(2, 6, 29)

#define SX16_KERNEL
#include "decomp_six.h"
#include "sx16core.h"

#include "sx16mod.h"
#include "sx16ioctl.h"

#define RESOURCE_VIDEO 1


MODULE_LICENSE("GPL");

#define SX16_MAJOR_VERSION	2
#define SX16_MINOR_VERSION	39
#define SX16_RELEASE		0
#define SX16_VERSION		KERNEL_VERSION(SX16_MAJOR_VERSION,	\
					       SX16_MINOR_VERSION,	\
					       SX16_RELEASE)


#define ERESTARTX16 519
#define MAX_WIDTH   720


static int debug;
static int *sx16_debug = &debug;

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 5, 0)
module_param(debug, int, 0);
MODULE_PARM_DESC(debug, "Debug level(0-100) default 0");
#endif

#define dprintk(level, fmt, arg...)				\
	do {							\
		if (*sx16_debug >= (level)) {			\
			printk(KERN_DEBUG "sx16: " fmt, ##arg);	\
		}						\
	} while (0)


#if defined(SX16_V4L_BUILD)
static int vid_limit = 16;

struct sx16_fmt {
	char *name;
	u32 fourcc;
	int depth;
};

/* sx16 page table */
struct sx16_pgtable {
	unsigned int	size;
	u32		*cpu;
	dma_addr_t	dma;
	/* used for offsets for u,v planes for planar capture modes */
	unsigned long	offset;
	/* used for custom pagetables (used for example by budget dvb cards) */
	struct scatterlist *slist;
};

struct sx16_buffer {
	/* common v4l buffer stuff -- must be first */
	struct videobuf_buffer		vb;
	const struct sx16_fmt		*fmt;
	int				tvnorm;
	struct sx16_pgtable		pt;
	int				btformat;
	int				btswap;
};

struct sx16_dmaqueue{
	/* back reference to board/core*/
	int		    board;
	HBD		    hbd;
	struct sx16_v4l	    *v4l;
	struct sx16_buffer  *curr;
	struct list_head    queue;
	struct timer_list   timeout;
};

struct sx16_fh {
	struct sx16_v4l		    *v4l;
	int			    resources;
	int			    board;
#ifdef VIDIOC_G_PRIORITY
	enum v4l2_priority	    prio;
#endif
	enum v4l2_buf_type	    type;
	/* capture */
	struct videobuf_queue	    cap;
	const struct sx16_fmt	    *fmt;
	int			    width;
	int			    height;
	int			    depth;
	int			    lines;
};


struct sx16_v4l {
#if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,29)
	struct v4l2_device v4l2_dev;
#endif
	struct pci_dev *pci;
	struct sx16_fh init;
#ifdef VIDIOC_G_PRIORITY
	struct v4l2_prio_state prio;
#endif
	/* locking */
	spinlock_t slock;
	/* back reference to board/core */
	int		     board;

	HBD		     hbd;
	struct sx16_dmaqueue video_q;
	int		     resources;
	struct sx16_fh	    *video_fh; /* back reference */
	struct mutex	     lock;
	int is_ntsc;
};

#endif
#if defined(SX16_V4L_BUILD)
static inline struct sx16_v4l *to_sx16_v4l(struct v4l2_device *v4l2_dev)
{
	return container_of(v4l2_dev, struct sx16_v4l, v4l2_dev);
}
#endif
#if LINUX_VERSION_CODE >= KERNEL_VERSION(4, 11, 0)
#include <linux/sched/signal.h>
#define wait_queue_t wait_queue_entry_t
#endif


#if LINUX_VERSION_CODE >= KERNEL_VERSION(5, 0, 0 )
#define ACCESS_OK(x, y, z) access_ok(y, z)

void do_gettimeofday(struct timeval *tv)
{
	struct timespec64 now;

	ktime_get_real_ts64(&now);	
	tv->tv_sec = now.tv_sec;
	tv->tv_usec = now.tv_nsec/1000;
}

#else
#define ACCESS_OK(x, y, z) access_ok(x, y, z)
#endif


#define __poll_event_interruptible_timeout(wq, condition, ret)	\
	({							\
		wait_queue_t __wait;				\
		init_waitqueue_entry(&__wait, current);		\
								\
		add_wait_queue(&wq, &__wait);			\
		for (; ; ) {					\
			set_current_state(TASK_INTERRUPTIBLE);	\
			if (condition || ret == 0)		\
				break;				\
			if (!signal_pending(current)) {		\
				schedule_timeout(1);		\
				ret--;				\
				continue;			\
			}					\
			ret = -ERESTARTSYS;			\
			break;					\
		}						\
		current->state = TASK_RUNNING;			\
		remove_wait_queue(&wq, &__wait);		\
	})

#define poll_event_interruptible_timeout(wq, condition, timeout)	\
	({								\
		int __ret = timeout;					\
		if (!(condition))					\
			__poll_event_interruptible_timeout(wq, \
							   condition, __ret); \
		__ret;							\
	})

#define __poll_event_interruptible_timeout_test(bd, wq, condition, ret, timeout2) \
	({								\
		wait_queue_t __wait;					\
		int __oldql = CORE_queue_get_any_size(bd, Q_ID_MUX); \
		int __ql;						\
		int __qcnt = 0;						\
		init_waitqueue_entry(&__wait, current);			\
									\
		add_wait_queue(&wq, &__wait);				\
		for (; ; ) {						\
			set_current_state(TASK_INTERRUPTIBLE);		\
			if (condition || ret == 0)			\
				break;					\
			if (!signal_pending(current)) {			\
				schedule_timeout(1);			\
				ret--;					\
				__ql = CORE_queue_get_any_size(bd, \
								   Q_ID_MUX); \
				if (__ql == __oldql) {		\
					__qcnt++;			\
					if (__qcnt > timeout2) {	\
						condition;		\
						ret = -ERESTARTX16;	\
						break;			\
					}				\
				} else {				\
					__oldql = __ql;			\
					__qcnt = 0;			\
				}					\
				continue;				\
			}						\
			ret = -ERESTARTSYS;				\
			break;						\
		}							\
		current->state = TASK_RUNNING;				\
		remove_wait_queue(&wq, &__wait);			\
	})

#define poll_event_interruptible_timeout_test(bd, wq, condition, timeout1, timeout2) \
	({								\
		int __ret = timeout1;					\
		if (!(condition))					\
			__poll_event_interruptible_timeout_test(bd, wq, \
								condition, \
								__ret, \
								timeout2);\
		__ret;	\
	})


/* Device name for sx16 v4l interface*/

#define DP printk(KERN_INFO "sx16mod:%d\n", __LINE__)
#define VIDEO_PREFIX "video"

/* Manifest constants ---------------------------------*/
#define TRUE				1
#define FALSE				0

/* Size of contiguous I/O space occupied by the 516 board.*/
#define REGION_SIZE_516		32
/* Size of SAA7146 application memory-mapped I/O space (PCI region 0)*/
/* occupied by the 616 board.*/
#define REGION_SIZE_616		512
/* For linux, the practical minimum sleep time is about 10ms.*/
#define MIN_SLEEP_TIME		10
/* Module/board installation success code.*/
#define INSTALL_PASS		0
/* Module/board installation fail code. */
#define INSTALL_FAIL		-1


/* Configuration constants ---------------------------- */
#define START_TIMEOUT		(HZ * 3)
/* Wait timeout for stream read/write, in milliseconds.*/
#define SLEEP_TIMEOUT_MSEC	10000
/* Wait timeout for stream read/write when "non-blocking", in ms */
#define NOBLK_TIMEOUT_MSEC	1000
#define HIGH_WATERMARK_VIDEO		 65535	/* QM, QV 20 bits*/
#define HIGH_WATERMARK_AUDIO		0xc000	/* QA 16 bits*/
#define LOW_WATERMARK	    0x0800
#define MAX_BOARDS			16
#define MAX_DEVS			32

/* Default values that are applied in place of missing insmod parameters:*/

/* Major device number.*/
#define DFLT_MAJOR		126
/* Default size of mpeg data used by poll to determine read/write readiness.*/
#define DFLT_POLL_SIZE		4096
/* Video bit rate in bps.*/
#define DFLT_BIT_RATE		4000000
/* MPEG multiplex format:  STREAMTYPE_MUX | STREAMTYPE_VES | STREAMTYPE_AES.*/
#define DFLT_STREAM_TYPE	STREAMTYPE_MUX
/* PictureSize_mpegType.*/
#define DFLT_MPEG			SIZE704_2
/* Input type: CVIDEO (composite) or SVIDEO.*/
#define DFLT_INPUT_TYPE		CVIDEO
/* Input number: 0 to 3.*/
#define DFLT_INPUT_NUMBER	0
/* Video system: SYS_NTSC or SYS_PAL.*/
#define DFLT_VIDEO_SYSTEM	SYS_NTSC
/* Variable bit rate enable.*/
#define DFLT_VBR_ENABLED	FALSE
/* Audio volume: 0 = maximum, 63 = minimum.*/
#define DFLT_VOLUME			0
/* Bass level: 0 = maximum.*/
#define DFLT_BASS			0
/* Treble level: 0 = maximum.*/
#define DFLT_TREBLE			0
/* Mute enable.*/
#define DFLT_MUTE			FALSE
/* Audio format.*/
#define DFLT_AUDIO_FORMAT	AFMT_MPEG1_LAYER2
/* Stereo mode.*/
#define DFLT_STEREO_MODE	STEREOMODE_STEREO
/* Mux A/V select: AVSEL_BOTH, AVSEL_VIDEO_ONLY, AVSEL_AUDIO_ONLY.*/
#define DFLT_AV_SELECT		AVSEL_BOTH
/* Luminance brightness = above nominal (128).*/
#define DFLT_BRIGHTNESS		128
/* Luminance contrast = above nominal (64).*/
#define DFLT_CONTRAST		68
/* Chrominance saturation = nominal.*/
#define DFLT_SATURATION		64
/* Audio sample rate = 48 KHz. */
#define DFLT_AUDIO_SMPRATE  AUDIO_SAMPLERATE_48KHZ
/* PCM resolution = 16 bits.*/
#define DFLT_AUDIO_PCM_RES  AUDIO_PCM_RES_16
/* Audio bit rate = 224 kbps (assumes audio format = MPEG-1 layer 2).*/
#define DFLT_AUDIO_BITRATE	11

/* GOP size.*/
#define DFLT_GOP			15
/* IP distance.*/
#define DFLT_IPDIST			3
/* default GPIO*/
#define DFLT_GPIO			3
/* audio AGC off*/
#define DFLT_AGC_ONOFF	    DFLT_AUDIO_AGC_ONOFF
/* audio AGC microphone sensitivity if AGC on 0dB*/
#define DFLT_AGC_SENS	    DFLT_AUDIO_AGC_SENS
/* audio input amplifier gain 0dB*/
#define DFLT_INAMPGAIN	    DFLT_AUDIO_INAMPGAIN

/* globals */
#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 5, 0)
/* Polling thread is running.*/
static int poll_thread_exists[MAX_BOARDS];
/* Thread Control Block for the polling thread.*/
static kthread_t poll_thread_tcb[MAX_BOARDS];

/* Thread Control Block for the SM start thread.*/
static kthread_t start_thread_tcb[MAX_BOARDS];
#else
/* Thread tast structure for the polling thread.*/
static struct task_struct *poll_thread_tcb[MAX_BOARDS];
/* Thread Control B*/
static struct task_struct   *start_thread_tcb[MAX_BOARDS];
#endif

/* flag used to check if thread started*/
static int start_thread_flag[MAX_BOARDS];
static char poll_thread_name[MAX_BOARDS][64];
static char start_thread_name[MAX_BOARDS][64];

/* Board handles*/
static HBD	hbd[MAX_BOARDS];
/* Board count obtained from insmod parameter count or pci scan.*/
static int	num_boards;
/* Model number to use for installation; assume 516 initially.*/
static int	model_num = 516;

/* Pointer to pci_dev structures for model 616 boards.*/
static struct pci_dev		*s616_dev		[MAX_BOARDS];
/* Wait queues for processes that are waiting to read/write stream data.*/
static wait_queue_head_t	stream_q		[MAX_BOARDS];
/* Wait queues for processes that are waiting to video field/frame sync.*/
static wait_queue_head_t	video_q		[MAX_BOARDS];
static int board_installed[MAX_BOARDS];
/* Logical base addresses of application registers.*/
static void *log_base[MAX_BOARDS];
static int open_count_v4l[MAX_BOARDS];
static int open_count_mpeg[MAX_BOARDS];
static int write_count[MAX_BOARDS];
/* Array of board numbers used to pass board number to interrupt handler.*/
static int board_number	[MAX_BOARDS];
/* Array of device names.*/
#define DEVNAME_ROOT "sx16a"
static char devname[MAX_BOARDS][64];

/* Pointers to 616 dma buffers.*/
static void *dmabuf[MAX_BOARDS];

static unsigned long PacketCountRx[MAX_BOARDS];

#ifdef AUDIO_EXTRA_FMTS_AC3
/* SM2288 encoder microcode.*/
extern unsigned char record_microcode_600[];

#endif

#ifdef AUDIO_EXTRA_FMTS_MP3
/* SM2288 encoder microcode.*/
extern unsigned char record_microcode_083[];
#endif
/* SM2288 encoder microcode.*/
extern unsigned char		record_microcode[];
/* SM2288 encoder microcode transport stream.*/
extern unsigned char		record_microcode_ts[];
/* SM2288 decoder microcode.*/
extern unsigned char		play_microcode	[];
/* SM2288 decoder microcode.*/
extern unsigned char		play_microcode_ves  [];

static struct semaphore sx16_devices_lock;

#if defined(SX16_V4L_BUILD)

/* video 4 linux properties*/
#define NUM_V4L2_STDS 2
#define V4L2_IDX_NTSC 0
#define V4L2_IDX_PAL  1



static int v4l2_to_sray_std(int id)
{
	if (id & V4L2_STD_NTSC)
		return SYS_NTSC;
	else if (id & V4L2_STD_PAL)
		return SYS_PAL;
	else
		return -1;
}

static const struct v4l2_queryctrl no_ctrl = {
	.name  = "v4l2",
	.flags = V4L2_CTRL_FLAG_DISABLED,
};

static const struct v4l2_queryctrl video_ctrls[] = {
	{
		.id	       = V4L2_CID_BRIGHTNESS,
		.name	       = "Brightness",
		.minimum       = 0,
		.maximum       = 255,
		.step	       = 1,
		.default_value = DFLT_BRIGHTNESS,
		.type	       = V4L2_CTRL_TYPE_INTEGER,
	}, {
		.id	       = V4L2_CID_CONTRAST,
		.name	       = "Contrast",
		.minimum       = 0,
		.maximum       = 127,
		.step	       = 1,
		.default_value = DFLT_CONTRAST,
		.type	       = V4L2_CTRL_TYPE_INTEGER,
	}, {
		.id	       = V4L2_CID_SATURATION,
		.name	       = "Saturation",
		.minimum       = 0,
		.maximum       = 127,
		.step	       = 1,
		.default_value = DFLT_SATURATION,
		.type	       = V4L2_CTRL_TYPE_INTEGER,
	}
};

static const struct sx16_fmt formats[] = {
	{
		.name = "RGB 24",
		.fourcc = V4L2_PIX_FMT_BGR24,
		.depth = 24

	}, {
		.name = "RGB 32",
		.fourcc = V4L2_PIX_FMT_BGR32,
		.depth = 32

	}, {
		.name = "RGB 565",
		.fourcc = V4L2_PIX_FMT_RGB565,
		.depth = 16
	}, {
		.name = "8bpp GREY",
		.fourcc = V4L2_PIX_FMT_GREY,
		.depth = 8
	}
};




static struct sx16_v4l *G_v4l[MAX_BOARDS];

static void sx16_dma_free(struct videobuf_queue *q, struct sx16_buffer *buf);

/* for V4L1 */
void sx16_pgtable_free(struct pci_dev *pci, struct sx16_pgtable *pt)
{
	if (NULL == pt->cpu)
		return;
	pci_free_consistent(pci, pt->size, pt->cpu, pt->dma);
#if 0
	dma_free_coherent(&pci->dev, pt->size,
			  &pt->cpu, pt->dma);
#endif
	pt->cpu = NULL;
	if (NULL != pt->slist) {
		kfree(pt->slist);
		pt->slist = NULL;
	}
}

#define PAGE_TABLE_SIZE_LOCAL (4096)

int sx16_pgtable_alloc(struct pci_dev *pci, struct sx16_pgtable *pt)
{
	u32	     *cpu;
	dma_addr_t   dma_addr;

	if (pt == NULL)
		return -ENOMEM;

	if (pci == NULL)
		return -ENOMEM;

	cpu = pci_alloc_consistent(pci, PAGE_TABLE_SIZE_LOCAL, &dma_addr);

	if (NULL == cpu)
		return -ENOMEM;

	pt->size = PAGE_TABLE_SIZE_LOCAL;
	pt->cpu	 = cpu;
	pt->dma	 = dma_addr;

	return 0;
}

int sx16_pgtable_build(struct pci_dev *pci, struct sx16_pgtable *pt,
		       struct scatterlist *list, int sglen)
{
	u32   *ptr, fill;
	int   nr_pages = 0;
	int   i, p;

	if (sglen == 0)	{
		dprintk(2, "sglen = 0: pgtablebuild. ERR.\n");
		return -EINVAL;
	}

	if (list->offset > PAGE_SIZE) {
		dprintk(2, "offset > PAGE_SIZE. this should not happen.");
		return -EINVAL;
	}
	/*
	 * if we have a user buffer, the first page may not be
	 *  aligned to a page boundary.
	 */
	pt->offset = list->offset;
	ptr = pt->cpu;
	for (i = 0; i < sglen; i++, list++) {
		for (p = 0; p * 4096 < list->length; p++, ptr++) {
			*ptr = sg_dma_address(list) + p * 4096;
			nr_pages++;
		}
	}
	/* required; fill the page table up with the last valid page */
	fill = *(ptr-1);

	for (i = nr_pages; i < 1024; i++)
		*ptr++ = fill;

	return 0;
}

/* updates the current DMA page table in the RPS program*/
int sx16_set_dma(struct sx16_v4l *v4l, struct sx16_buffer *buf,
		 struct sx16_buffer *next)
{
	int prot_addr;
	int base_even;
	int base_odd;
	int vflip = 0;
	int lines;
	int pitch;
	int bytesperline;
	int field;
	SX_DMA *dma = NULL;
	HBD h_hbd;
	if (v4l == NULL) {
		printk(KERN_ERR "%s invalid v4l\n", __func__);
		return -1;
	}

	h_hbd = v4l->hbd;
	if (h_hbd == NULL) {
		printk(KERN_ERR "%s invalid handle\n", __func__);
		return -1;
	}
	CORE_get_dma_vformat(h_hbd, &dma, NULL);

	if (dma == NULL)
		return -1;

	if (buf == NULL)
		return -1;

	if (buf->fmt == NULL)
		return -1;

	field = buf->vb.field;
	lines = buf->vb.height;
	bytesperline = buf->vb.width * (buf->fmt->depth >> 3);
	dprintk(4, "bytesperline %d w %d h %d depth %d\n",
		bytesperline, buf->vb.width, lines,
		buf->fmt->depth >> 3);

	pitch = bytesperline*2;

	if (0 != vflip) {
		dprintk(2, "vflipped\n");
		pitch *= -1;
		prot_addr = buf->pt.offset;
		base_even = buf->pt.offset - (pitch/2) * lines;
		base_odd = base_even + (pitch/2);
	} else {
		base_even = buf->pt.offset;
		base_odd = buf->pt.offset + (pitch/2);
		prot_addr = buf->pt.offset + (pitch/2) * (lines - 2);
	}
	/*  update for separate TOP and BOTTOM fields)*/
	if (field == VIDCAP_FIELD_ALTERNATE) {
		printk(KERN_INFO "vidcap: field alternate: not supported\n");
		/* one dword past protection addresss*/
		base_odd = prot_addr + (pitch > 0) ? 4 : -4;
		pitch /= 2;
	} else if (field == VIDCAP_FIELD_TOP) {
		/* one dword past protection addresss*/
		base_odd = prot_addr + (pitch > 0) ? 4 : -4;
		pitch /= 2;
	} else if (field == VIDCAP_FIELD_BOTTOM) {
		base_odd = base_even;
		/* one dword past protection addresss*/
		base_even = prot_addr + (pitch > 0) ? 4 : -4;
		pitch /= 2;
	}

	if (dma->rps_base_odd)
		*dma->rps_base_odd  = base_odd;

	if (dma->rps_base_even)
		*dma->rps_base_even = base_even;

	if (dma->rps_prot_addr)
		*dma->rps_prot_addr = prot_addr;

	if (dma->rps_pitch)
		*dma->rps_pitch = pitch;

	if (dma->rps_ptphys) {
#define DMA_MAP			0x800		/*DMA mapping enable*/
		/* set the physical page table address for the RPS program*/
		*dma->rps_ptphys = buf->pt.dma | DMA_MAP;
	}
	return 0;
}


static int buffer_activate(struct sx16_v4l *v4l,
			    struct sx16_buffer *buf,
			    struct sx16_buffer *next)
{

	if (buf == NULL)
		return -EINVAL;

	buf->vb.state = VIDEOBUF_ACTIVE;

	/* set the dma address for this RPS program.*/
	sx16_set_dma(v4l, buf, next);
	/* start the RPS again*/
	CORE_capture_start(v4l->hbd, 0);
	/* we should get a frame in at least 400 ms */
	mod_timer(&v4l->video_q.timeout, jiffies + msecs_to_jiffies(400));
	return 0;
}



static int buffer_prepare(struct videobuf_queue *q, struct videobuf_buffer *vb,
			  enum v4l2_field field)

{
	struct sx16_fh *fh = q->priv_data;
	struct sx16_buffer *buf = (struct sx16_buffer *) vb;
	struct sx16_v4l *v4l;
	int rc = 0;
	struct videobuf_dmabuf *dma;
	dprintk(2, "sx16 prepare buffer: vbuf:%p\n", vb);

	if (fh == NULL) {
		printk(KERN_INFO "sx16: invalid fh\n");
		return -EINVAL;
	}

	if (buf == NULL)
		return -EINVAL;

	v4l = fh->v4l;

	/* sanity checks */
	if (fh->width  < 48 ||
	    fh->height < 32 ||
	    fh->width  > MAX_WIDTH ||
	    fh->height > 576) {
		printk(KERN_INFO "invalid buffer size\n");
		return -EINVAL;
	}

	if (fh->fmt == NULL) {
		printk(KERN_INFO "invalid buffer format\n");
		return -EINVAL;
	}


	buf->vb.size = fh->width * fh->height * (fh->fmt->depth >> 3);

	if (0 != buf->vb.baddr && buf->vb.bsize < buf->vb.size) {
		dprintk(4, "invalid buffer prepare\n");
		return -EINVAL;
	}

	if (buf->fmt != fh->fmt ||
	    buf->vb.width != fh->width ||
	    buf->vb.height != fh->height || buf->vb.field != field) {
		buf->fmt = fh->fmt;
		buf->vb.width = fh->width;
		buf->vb.height = fh->height;
		buf->vb.field = field;
		sx16_dma_free(q, buf);
	}

	buf->fmt = fh->fmt;

	if (VIDEOBUF_NEEDS_INIT == buf->vb.state) {
		/* build and allocate the page table*/
		if (sx16_pgtable_alloc(v4l->pci, &buf->pt) != 0) {
			printk(KERN_INFO "%s: alloc failed\n", __func__);
			goto fail;
		}

		rc = videobuf_iolock(&fh->cap, &buf->vb, NULL);

		if (rc) {
			printk(KERN_INFO "%s iolock fail %d\n", __func__, rc);
			sx16_pgtable_free(v4l->pci, &buf->pt);
			goto fail;
		}



		/* build the page table*/
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 24)
		dma = videobuf_to_dma(&buf->vb);
#else
		dma = &buf->vb.dma;
#endif

		if (dma != NULL) {
			rc = sx16_pgtable_build(v4l->pci, &buf->pt,
						dma->sglist, dma->sglen);
		} else {
			printk(KERN_INFO "could not get dma address\n");
			sx16_pgtable_free(v4l->pci, &buf->pt);
			goto fail;
		}

		if (rc) {
			printk(KERN_INFO "%s failed\n", __func__);
			sx16_pgtable_free(v4l->pci, &buf->pt);
			goto fail;
		}
		buf->vb.state = VIDEOBUF_PREPARED;
	}
	return 0;
fail:
	sx16_dma_free(q, buf);
	return rc;
}

static int sx16_buffer_pages(int size)
{
	size  = PAGE_ALIGN(size);
	/* for non-page-aligned buffers */
	size += PAGE_SIZE;
	size /= 4096;
	return size;
}

int sx16_buffer_count(unsigned int size, unsigned int count)
{
	unsigned int maxcount;

	maxcount = 1024 / sx16_buffer_pages(size);
	if (count > maxcount)
		count = maxcount;
	return count;
}

static int buffer_setup(struct videobuf_queue *vq, unsigned int *count,
			unsigned int *size)
{
	struct sx16_fh *fh = vq->priv_data;


	*size = fh->width * fh->height * (fh->fmt->depth >> 3);

	if (0 == *count)
		*count = vid_limit;
	*count = sx16_buffer_count(*size, *count);
	return 0;
}


static void sx16_dma_free(struct videobuf_queue *q, struct sx16_buffer *buf)
{
	struct videobuf_dmabuf *dma;
	if (buf == NULL)
		return;
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 37))
	videobuf_waiton(q, &buf->vb, 0, 0);
#else
	videobuf_waiton(&buf->vb, 0, 0);
#endif
	
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 24)
	dma = videobuf_to_dma(&buf->vb);
#else
	dma = &buf->vb.dma;
#endif

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 36))	
	videobuf_dma_unmap(q->dev, dma);
#else
	videobuf_dma_unmap(q, dma);
#endif
	
	videobuf_dma_free(dma);

	buf->vb.state = VIDEOBUF_NEEDS_INIT;
	return;
}



static void buffer_queue(struct videobuf_queue *vq, struct videobuf_buffer *vb)
{
	struct sx16_buffer *buf = (struct sx16_buffer *)vb;
	struct sx16_fh *fh = vq->priv_data;
	struct sx16_dmaqueue *q;
	struct sx16_v4l *v4l;/* = fh->v4l;*/

	if (fh == NULL) {
		printk(KERN_INFO "q err\n");
		return;
	}

	v4l = fh->v4l;
	if (v4l == NULL) {
		printk(KERN_INFO "v4l err\n");
		return;
	}

	q = &fh->v4l->video_q;
	if (q == NULL) {
		printk(KERN_INFO "%s: v4l err\n", __func__);
		return;
	}

	if (NULL == q->curr) {
		q->curr = buf;
		buffer_activate(v4l, buf, NULL);
	} else {
		list_add_tail(&buf->vb.queue, &q->queue);
		buf->vb.state = VIDEOBUF_QUEUED;
	}

	return;
}


static void buffer_release(struct videobuf_queue *vq,
			   struct videobuf_buffer *vb)
{
	struct sx16_buffer   *buf = (struct sx16_buffer *) vb;
	struct sx16_fh *fh = vq->priv_data;
	struct sx16_v4l	     *v4l;

	dprintk(4, "buffer release %p \n", buf);
	if (fh) {
		v4l = fh->v4l;
		if (v4l == NULL)
			return;
		sx16_dma_free(vq, buf);
	}
}

void sx16_buffer_finish(struct sx16_v4l *dev,
			struct sx16_dmaqueue *q,
			int state)
{
#if LINUX_VERSION_CODE >= KERNEL_VERSION(5, 0, 0 ) && LINUX_VERSION_CODE < KERNEL_VERSION(5, 4, 0)
	struct timespec64 ts;
#endif
	assert_spin_locked(&dev->slock);
	if (q == NULL) {
		printk(KERN_INFO "q null\n");
		return;
	}
	if (NULL == q->curr) {
		printk(KERN_INFO "internal error: NULL pointer for curr.\n");
		return;
	}

#ifdef DEBUG_VIDEOBUF
	printk(KERN_DEBUG "dev:%p, dmaq:%p, state:%d\n", dev, q, state);
	printk(KERN_DEBUG"q->curr:%p\n", q->curr);
#endif

	/* finish current buffer */
	if (NULL == q->curr) {
		dprintk(4, "no current buffer\n");
		return;
	}
	q->curr->vb.state = state;
#if LINUX_VERSION_CODE < KERNEL_VERSION(5, 0, 0 )
	do_gettimeofday(&q->curr->vb.ts);
#else
#if LINUX_VERSION_CODE < KERNEL_VERSION(5, 4, 0)
	ktime_get_real_ts64(&ts);
	q->curr->vb.ts.tv_sec = ts.tv_sec;
	q->curr->vb.ts.tv_usec = ts.tv_nsec / NSEC_PER_USEC;
#else
	// TODO. update time for 5.4+
#endif
#endif
	
	wake_up(&q->curr->vb.done);
	q->curr = NULL;
}

void sx16_buffer_next(struct sx16_v4l *v4l, struct sx16_dmaqueue *q)
{
	struct sx16_buffer *buf, *next = NULL;

	assert_spin_locked(&v4l->slock);
	if (NULL == q)
		return;

	if (!list_empty(&q->queue)) {
		buf = list_entry(q->queue.next, struct sx16_buffer, vb.queue);
		list_del(&buf->vb.queue);
		if (!list_empty(&q->queue)) {
			next = list_entry(q->queue.next,
					  struct sx16_buffer,
					  vb.queue);
		}
		q->curr = buf;
		dprintk(4, "next buffer: buf:%p, prev:%p, next:%p\n",
			buf, q->queue.prev, q->queue.next);

		buffer_activate(v4l, buf, next);
	} else {
		del_timer(&q->timeout);
		CORE_capture_stop(q->hbd);
		dprintk(4, "buffer next no queue\n");
	}
	return;
}

//	struct sx16_dmaqueue video_q;
//	setup_timer(&v4l->video_q.timeout, sx16_buffer_timeout, (unsigned long) &v4l->video_q);
#if LINUX_VERSION_CODE >= KERNEL_VERSION(4, 8, 0)
static void sx16_buffer_timeout(struct timer_list *t)
{
	struct sx16_dmaqueue *q = from_timer(q, t, timeout);
#else
static void sx16_buffer_timeout(unsigned long data)
{
	struct sx16_dmaqueue *q = (struct sx16_dmaqueue *) data;
#endif
	struct sx16_v4l *v4l;
	unsigned long flags;
	dprintk(4, "sx16: =========buffer_timeout===========\n");
	if (q == NULL)
		return;
	v4l = q->v4l;
	if (v4l == NULL)
		return;

	spin_lock_irqsave(&v4l->slock, flags);
	if (q->curr) {
#ifdef DEBUG_VIDEOBUF
		printk(KERN_INFO "MC1:%x\n", CORE_capture_get_MC1(q->hbd));
		printk(KERN_INFO "MC2:%x\n", CORE_capture_get_MC2(q->hbd));
		printk(KERN_INFO "PSR:%x\n", CORE_capture_get_PSR(q->hbd));
		printk(KERN_INFO "SSR:%x\n", CORE_capture_get_SSR(q->hbd));
#endif
		sx16_buffer_finish(v4l, q, VIDEOBUF_ERROR);
	}
	sx16_buffer_next(v4l, q);
	spin_unlock_irqrestore(&v4l->slock, flags);
}

static void video_irq_done(struct sx16_v4l *v4l, unsigned long st)
{
	struct sx16_dmaqueue *q;

	if (v4l == NULL)
		return;

	spin_lock(&v4l->slock);
	q = &v4l->video_q;

	if (NULL != q->curr)
		sx16_buffer_finish(v4l, q, VIDEOBUF_DONE);

	sx16_buffer_next(v4l, q);
	spin_unlock(&v4l->slock);
}


int res_get(struct sx16_fh *fh, unsigned int bit)
{
	struct sx16_v4l *v4l = fh->v4l;

	if (fh->resources & bit)
		return 1;

	mutex_lock(&v4l->lock);
	if (v4l->resources & bit) {
		mutex_unlock(&v4l->lock);
		return 0;
	}
	fh->resources  |= bit;
	v4l->resources |= bit;
	mutex_unlock(&v4l->lock);
	return 1;
}

int res_check(struct sx16_fh *fh, unsigned int bit)
{
	return (fh->resources & bit);
}

int res_locked(struct sx16_v4l *v4l, unsigned int bit)
{
	return (v4l->resources & bit);
}

void res_free(struct sx16_fh *fh, unsigned int bits)
{
	struct sx16_v4l *v4l = fh->v4l;

	if ((fh->resources & bits) != bits)
		return;
	mutex_lock(&v4l->lock);
	fh->resources  &= ~bits;
	v4l->resources &= ~bits;
	mutex_unlock(&v4l->lock);
}





static struct videobuf_queue_ops sx16_video_qops = {
	.buf_setup    = buffer_setup,
	.buf_prepare  = buffer_prepare,
	.buf_queue    = buffer_queue,
	.buf_release  = buffer_release,
};


#endif /* #if defined(SX16_V4L_BUILD)*/


/*DEBUG*/
/*#define DBG*/
#ifdef DBG
#define DBGDATA_SIZE		(2 * 1024 * 1024)
struct DBGDATA {
	int			flag;
	int			bytes;
	char		data[DBGDATA_SIZE];
}

static struct DBGDATA		dbgdata;

void SX16MOD_dbg_wr_dword(unsigned long dw)
{
	static int flag = 0;
	if (dbgdata.bytes + 4 < DBGDATA_SIZE) {
		*(unsigned long *) &(dbgdata.data[dbgdata.bytes]) = dw;
		dbgdata.bytes += 4;
	} else if (!flag) {
		printk(KERN_DEBUG "Debug full\n\n");
		flag = 1;
	}
}

#endif /*DBG*/

/* Allocate storage for option insmod parameters
   This are initialized in init_module to the correct defaults. */
/* Major device number.*/
static unsigned int major = DFLT_MAJOR;
/* Physical base I/O address of application registers.*/
static unsigned long io[MAX_BOARDS] = {[0 ... (MAX_BOARDS-1)] = 0};
/* IRQ line assigned to board.*/
static unsigned int irq[MAX_BOARDS] = {[0 ... (MAX_BOARDS-1)] = 0};
/* Input type (composite, svideo).*/
static unsigned short intype[MAX_BOARDS] = {[0 ... (MAX_BOARDS-1)]
					    = DFLT_INPUT_TYPE};
/* Input number (0 - 3).*/
static unsigned short	inchan[MAX_BOARDS] = {[0 ... (MAX_BOARDS-1)]
					      = DFLT_INPUT_NUMBER};
/* Video system (PAL, NTSC).*/
static unsigned short system[MAX_BOARDS] = {[0 ... (MAX_BOARDS-1)]
					    = DFLT_VIDEO_SYSTEM};
/* Video bit rate in bps.*/
static unsigned long vbitrate[MAX_BOARDS] = {[0 ... (MAX_BOARDS-1)]
					     = DFLT_BIT_RATE};
/* Enumerated picture size/mpeg type.*/
static unsigned short mpeg[MAX_BOARDS] = {[0 ... (MAX_BOARDS-1)]
					  = DFLT_MPEG};
/* Variable bit rate compression enable.*/
static unsigned short vbron[MAX_BOARDS] = {[0 ... (MAX_BOARDS-1)]
					   = DFLT_VBR_ENABLED};
/* MPEG multiplex format (audioES, videoES, SVCD, VCD, ...).*/
static unsigned short streamtype[MAX_BOARDS] = {[0 ... (MAX_BOARDS-1)]
						= DFLT_STREAM_TYPE};
/* Audio volume.*/
static unsigned short vol[MAX_BOARDS] = {[0 ... (MAX_BOARDS-1)] = DFLT_VOLUME};
static unsigned short bass[MAX_BOARDS] = {[0 ... (MAX_BOARDS-1)] = DFLT_BASS};
/* Treble level.*/
static unsigned short treb[MAX_BOARDS] = {[0 ... (MAX_BOARDS-1)] = DFLT_TREBLE};
static unsigned short mute[MAX_BOARDS] = {[0 ... (MAX_BOARDS-1)] = DFLT_MUTE};
static unsigned short afmt[MAX_BOARDS] = {[0 ... (MAX_BOARDS-1)]
					  = DFLT_AUDIO_FORMAT};
/* Audio Sample rate.*/
static unsigned short asmprate[MAX_BOARDS] = {[0 ... (MAX_BOARDS-1)]
					      = DFLT_AUDIO_SMPRATE};
/* Stereo mode.*/
static unsigned short smode[MAX_BOARDS] = {[0 ... (MAX_BOARDS-1)]
					   = DFLT_STEREO_MODE};
/* AV select.*/
static unsigned short avsel[MAX_BOARDS] = {[0 ... (MAX_BOARDS-1)]
					   = DFLT_STEREO_MODE};
/* GOP size.*/
static unsigned short gop[MAX_BOARDS] = {[0 ... (MAX_BOARDS-1)]
					   = DFLT_GOP};
/* IP distance.*/
static unsigned short ipdist[MAX_BOARDS] = {[0 ... (MAX_BOARDS-1)]
					   = DFLT_IPDIST};
/* audio settings*/
/* AGC on off*/
static unsigned short agcon[MAX_BOARDS] = {[0 ... (MAX_BOARDS-1)]
					   = DFLT_AGC_ONOFF};
/* AGC sensitivity(AGC on only)*/
static unsigned short agcsens[MAX_BOARDS] = {[0 ... (MAX_BOARDS-1)]
					   = DFLT_AGC_SENS};
/* Input gain(AGC off only)*/
static unsigned short ampgain[MAX_BOARDS] = {[0 ... (MAX_BOARDS-1)]
					   = DFLT_INAMPGAIN};
static unsigned short gpio[MAX_BOARDS] = {[0 ... (MAX_BOARDS-1)]
					   = DFLT_GPIO};
/* nosubid: Set to 1 if there is no sub-id or subvendor-id on the 616s.*/
static unsigned int nosubid = 0;


#if defined(SX16_V4L_BUILD)
/* device numbers (video0, 1, ...) assigned for each board*/
static int devnum_v4l[MAX_BOARDS];
/* board numbers for each device*/
static int brdnum_v4l[MAX_DEVS];
static struct video_device *vdev[MAX_BOARDS];

static struct video_device *vdev_init(int board,
				      struct video_device *template,
				      char *type);


static int __devinit register_video(int board);
static void unregister_video(int board);
/* end V4l*/
#endif

static void video_sleepq(void *board, int timeout);

/* converts handle to board number -1 if none found */
static int handle_to_boardnum(HBD board)
{
	int board_num = -1;  /* board identifier */
	int i;
	i = 0;
	while (i < MAX_BOARDS) {
		if (hbd[i] == board) {
			board_num = i;
			break;
		}
		i++;
	}
	return board_num;
}


/* Number of packets to copy to user-space buffer.*/
#define NUM_SENSORAY_PACKETS	4
/* Packet header byte count: 4 (packet, incl. hdr.) + 2 (sequence num.).*/
#define SENSORAY_HEADER_SIZE	6
/* Number of mpeg bytes per packet.*/
#define SENSORAY_PAYLOAD_SIZE	1440
/* Number of mpeg bytes per packet.  transport stream read*/
#define SENSORAY_PAYLOAD_SIZE_TS 1316

static ssize_t read_sensoray(HBD board, char *buffer, int payloadsize);
#ifdef SX16_BUILD_SRAY
static ssize_t write_sensoray(HBD board, char *buffer, int skip_data);
static ssize_t write_sensoray_mux(HBD board, char *buffer, int skip_data);
#endif

#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 5, 0)
static void start_dec_thread(kthread_t *kthread);
#else
static int start_dec_thread(void *kthread);
#endif


static int start_decoder(int board, int skip_data);
#ifdef SX16_BUILD_SRAY
static int RTCP_SSCP_Report(HBD board, void *val, int cmd);
#endif
/* sx16_write_core encapsulates the main writing functions
 *    used in sx16_write and write_sensoray
 */

static ssize_t sx16_write_core(int board_num, const char *buffer,
			       size_t nbytes, int no_block)
{
	int rc;
	int bytes_written = 0;
	int ready;
	int out = 0;
	int bytes_desired;
	HBD h_hbd = hbd[board_num];

	/* If non-blocking and there is no space, return immediately.*/
	if (no_block && CORE_get_dma_remaining(h_hbd) < nbytes) {
		if ((CORE_queue_get_any_size(h_hbd, Q_ID_MUX)
		     > HIGH_WATERMARK_VIDEO) ||
		    (CORE_poll_dma_flag(h_hbd) == 0))
			return -EAGAIN;
	}

	/* Write the MPEG data to the SM...*/
	while (nbytes - bytes_written > 0) {
		bytes_desired = nbytes - bytes_written;
		/* Attempt to copy nbytes of MPEG data from user space.*/
		rc = CORE_stream_write(h_hbd,
				       (unsigned char *) &buffer[bytes_written],
				       bytes_desired);

		bytes_written += rc;

		/* Continue if we wrote the desired number of bytes.*/
		if (rc == bytes_desired) {
			out = 0;
			if (write_count[board_num] < 1024)
				write_count[board_num]++;
			continue;
		}

		/* Wait for dma flag. For each interrupt,
		   check condition until dma flag is true or timeout.
		   NOTE: This "non-blocking" call is a call
		   with a small timeout.*/
		rc = poll_event_interruptible_timeout_test(
			h_hbd, stream_q[board_num],
			(ready = (CORE_poll_dma_flag(h_hbd) && (CORE_queue_get_any_size(h_hbd, Q_ID_MUX) < HIGH_WATERMARK_VIDEO))),
			(((no_block ? NOBLK_TIMEOUT_MSEC : SLEEP_TIMEOUT_MSEC) * HZ) / 1000),
			(write_count[board_num] == 1024 ? HZ/20 : 100 * HZ));

		if (rc == -ERESTARTX16) {
			if (out++ < 10)
				continue;
		}
		/* If timeout, return number of bytes read*/
		if (!ready) {
			if ((no_block && bytes_written == 0) || !no_block)
				return -ENXIO;
			else
				break;
		}
	}

	return bytes_written;
}

/* call back function for sensoray write*/
unsigned long SX16MOD_write_core(HBD board, const char *buffer,
				  unsigned long nbytes, int no_block)
{
	int board_num;
	int rc;
	board_num = handle_to_boardnum(board);
	rc = sx16_write_core(board_num, buffer, nbytes, no_block);
	return rc;
}


/** sx16_read_core encapsulates the main reading functions
    used in sx16_read and read_sensoray
*/
static ssize_t sx16_read_core(int board_num, char *buffer, size_t nbytes,
			      int no_block)
{
	int	rc;
	int		bytes_read	= 0;
	int		bytes_desired;
	int	ready;

	HBD	h_hbd = hbd[board_num];

	/* Start the encoder if it is not already running.*/
	if (!CORE_stream_is_encoding(h_hbd)) {
		CORE_start_encode(h_hbd);
	}

	/* If non-blocking and there is no data, return immediately.*/
	if (no_block && CORE_get_dma_remaining(h_hbd) < nbytes) {
		if (CORE_queue_get_any_size(h_hbd, Q_ID_MUX) < LOW_WATERMARK
		     || CORE_poll_dma_flag(h_hbd) == 0) {
			return -EAGAIN;
		}
	}

	/* Read the MPEG data in the user buffer...*/
	while (nbytes - bytes_read > 0) {
		bytes_desired = nbytes - bytes_read;
		/* Attempt to copy nbytes of MPEG stream data to user space.*/
		rc = CORE_stream_read(h_hbd,
				      (unsigned char *) &buffer[bytes_read],
					  bytes_desired);
		bytes_read += rc;
		/* Loop if we read the desired number of bytes.*/
		if (rc == bytes_desired)
			continue;
		/* Otherwise, wait for dma flag. For each interrupt,
		   check condition until*/
		/* dma flag is true or timeout.
		   NOTE: This "non-blocking" call is a call
		   with a small timeout.*/
		poll_event_interruptible_timeout(stream_q[board_num],
		 (ready = CORE_poll_dma_flag(h_hbd)),
		 ((no_block ? NOBLK_TIMEOUT_MSEC : SLEEP_TIMEOUT_MSEC) * HZ)
		      / 1000);

		/* If we have a timeout, return.*/
		if (!ready) {
			if ((no_block && bytes_read == 0) || !no_block)
				return -ENXIO;
			else
				break;
		}
	}
	return bytes_read;
}

/* insmod parameters */
#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 5, 0)
MODULE_PARM(major, "i");
MODULE_PARM(io,    "1-" __MODULE_STRING(MAX_BOARDS) "i");
MODULE_PARM(irq,   "1-" __MODULE_STRING(MAX_BOARDS) "i");
MODULE_PARM(system, "1-" __MODULE_STRING(MAX_BOARDS) "h");
MODULE_PARM(intype, "1-" __MODULE_STRING(MAX_BOARDS) "h");
MODULE_PARM(inchan, "1-" __MODULE_STRING(MAX_BOARDS) "h");
MODULE_PARM(mpeg,  "1-" __MODULE_STRING(MAX_BOARDS) "h");
MODULE_PARM(vbron, "1-" __MODULE_STRING(MAX_BOARDS) "h");
MODULE_PARM(vbitrate, "1-" __MODULE_STRING(MAX_BOARDS) "l");
MODULE_PARM(gop,   "1-" __MODULE_STRING(MAX_BOARDS) "h");
MODULE_PARM(ipdist, "1-" __MODULE_STRING(MAX_BOARDS) "h");
MODULE_PARM(streamtype, "1-" __MODULE_STRING(MAX_BOARDS) "h");
MODULE_PARM(vol,   "1-" __MODULE_STRING(MAX_BOARDS) "h");
MODULE_PARM(bass,  "1-" __MODULE_STRING(MAX_BOARDS) "h");
MODULE_PARM(treb,  "1-" __MODULE_STRING(MAX_BOARDS) "h");
MODULE_PARM(mute,  "1-" __MODULE_STRING(MAX_BOARDS) "h");
MODULE_PARM(afmt,  "1-" __MODULE_STRING(MAX_BOARDS) "h");
MODULE_PARM(asmprate, "1-" __MODULE_STRING(MAX_BOARDS) "h");
MODULE_PARM(smode, "1-" __MODULE_STRING(MAX_BOARDS) "h");
MODULE_PARM(avsel, "1-" __MODULE_STRING(MAX_BOARDS) "h");
MODULE_PARM(gpio,  "1-" __MODULE_STRING(MAX_BOARDS) "h");
MODULE_PARM(ampgain, "1-" __MODULE_STRING(MAX_BOARDS) "h");
MODULE_PARM(agcon, "1-" __MODULE_STRING(MAX_BOARDS) "h");
MODULE_PARM(agcsens, "1-" __MODULE_STRING(MAX_BOARDS) "h");
MODULE_PARM(nosubid, "1-" __MODULE_STRING(MAX_BOARDS) "i");

#elif LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 10)
/* 2.6 kernel */
/*module_param_array changed from 2.6.9 to 2.6.10.
  See parameters below(value to pointer).  Have to
  ifdef for kernel version within the major kernel 2.6 release. */
static int sx16devs[23];
module_param(major, int, 0444);
module_param_array(io, int, sx16devs[0], 0444);
module_param_array(irq, int, sx16devs[1], 0444);
module_param_array(system, ushort, sx16devs[2], 0444);
module_param_array(intype, ushort, sx16devs[3], 0444);
module_param_array(inchan, ushort, sx16devs[4], 0444);
module_param_array(mpeg, ushort, sx16devs[5], 0444);
module_param_array(vbron, ushort, sx16devs[6], 0444);
module_param_array(vbitrate, long, sx16devs[7], 0444);
module_param_array(gop, ushort, sx16devs[8], 0444);
module_param_array(ipdist, ushort, sx16devs[9], 0444);
module_param_array(streamtype, ushort, sx16devs[10], 0444);
module_param_array(vol, ushort, sx16devs[11], 0444);
module_param_array(bass, ushort, sx16devs[12], 0444);
module_param_array(treb, ushort, sx16devs[13], 0444);
module_param_array(mute, ushort, sx16devs[14], 0444);
module_param_array(afmt, ushort, sx16devs[15], 0444);
module_param_array(asmprate, ushort, sx16devs[16], 0444);
module_param_array(smode, ushort, sx16devs[17], 0444);
module_param_array(avsel, ushort, sx16devs[18], 0444);
module_param_array(gpio, ushort, sx16devs[19], 0444);
module_param_array(ampgain, ushort, sx16devs[20], 0444);
module_param_array(agcon, ushort, sx16devs[21], 0444);
module_param_array(agcsens, ushort, sx16devs[22], 0444);
module_param(nosubid, int, 0444);
#else
/* 2.6.10 and over */
static int sx16devs[23];
module_param(major, int, 0444);
module_param_array(io, ulong, &sx16devs[0], 0444);
module_param_array(irq,	int, &sx16devs[1], 0444);
module_param_array(system, ushort, &sx16devs[2], 0444);
module_param_array(intype, ushort, &sx16devs[3], 0444);
module_param_array(inchan, ushort, &sx16devs[4], 0444);
module_param_array(mpeg, ushort, &sx16devs[5], 0444);
module_param_array(vbron, ushort, &sx16devs[6], 0444);
module_param_array(vbitrate, long, &sx16devs[7], 0444);
module_param_array(gop, ushort, &sx16devs[8], 0444);
module_param_array(ipdist, ushort, &sx16devs[9], 0444);
module_param_array(streamtype, ushort, &sx16devs[10], 0444);
module_param_array(vol,	ushort, &sx16devs[11], 0444);
module_param_array(bass, ushort, &sx16devs[12], 0444);
module_param_array(treb, ushort, &sx16devs[13], 0444);
module_param_array(mute, ushort, &sx16devs[14], 0444);
module_param_array(afmt, ushort, &sx16devs[15], 0444);
module_param_array(asmprate, ushort, &sx16devs[16], 0444);
module_param_array(smode, ushort, &sx16devs[17], 0444);
module_param_array(avsel, ushort, &sx16devs[18], 0444);
module_param_array(gpio, ushort, &sx16devs[19], 0444);
module_param_array(ampgain, ushort, &sx16devs[20], 0444);
module_param_array(agcon, ushort, &sx16devs[21], 0444);
module_param_array(agcsens, ushort, &sx16devs[22], 0444);
module_param(nosubid, int, 0444);
#endif


MODULE_PARM_DESC(major, "Driver major number (default="
		 __MODULE_STRING(DFLT_MAJOR) ")");
MODULE_PARM_DESC(io, "Base I/O address - not for PCI");
MODULE_PARM_DESC(irq, "IRQ select. ISA:3,5,7,10,11,15, or 0 to disable."
		 " PCI:-1 to disable (default=0)");
MODULE_PARM_DESC(system, "Video system 1-PAL 2-NTSC (default="
		 __MODULE_STRING(DFLT_VIDEO_SYSTEM) ")");
MODULE_PARM_DESC(intype, "Video input type 0-CVIDEO 1-SVIDEO (default="
		 __MODULE_STRING(DFLT_INPUT_TYPE)  ")");
MODULE_PARM_DESC(inchan, "Video input number 0-3 (default="
		 __MODULE_STRING(DFLT_INPUT_NUMBER) ")");
MODULE_PARM_DESC(mpeg, "Picture size/mpeg type 0-D1/MPEG2"
		 " 1-2/3D1/MPEG2 2-1/2D1/MPEG2 3-SIF/MPEG1 (default="
		 __MODULE_STRING(DFLT_MPEG)	    ")");
MODULE_PARM_DESC(vbron, "Variable bit rate 0-off 1-on (default="
		 __MODULE_STRING(DFLT_VBR_ENABLED) ")");
MODULE_PARM_DESC(vbitrate, "Video bit rate (default="
		 __MODULE_STRING(DFLT_BIT_RATE) ")");
MODULE_PARM_DESC(gop, "GOP size 0-31 (default="
		 __MODULE_STRING(DFLT_GOP)		   ")");
MODULE_PARM_DESC(ipdist, "IP distance 0-3 (default="
		 __MODULE_STRING(DFLT_IPDIST) ")");
MODULE_PARM_DESC(streamtype, "Stream type"
		 "0-mux 1-ves 2-aes 3-transport (default="
		 __MODULE_STRING(DFLT_STREAM_TYPE) ")");
MODULE_PARM_DESC(vol, "Audio volume 0-max 63-min (default="
		 __MODULE_STRING(DFLT_VOLUME) ")");
MODULE_PARM_DESC(bass, "Audio bass 0-max 63-min (default="
		 __MODULE_STRING(DFLT_BASS)	    ")");
MODULE_PARM_DESC(treb, "Audio Treble 0-max 63-min (default="
		 __MODULE_STRING(DFLT_TREBLE) ")");
MODULE_PARM_DESC(mute, "Audio mute 0-disable 1-enable (default="
		 __MODULE_STRING(DFLT_MUTE)	    ")");
MODULE_PARM_DESC(afmt, "Audio MPEG-1 layer 0-layer1 1-layer2 (default="
		 __MODULE_STRING(DFLT_AUDIO_FORMAT) ")");
MODULE_PARM_DESC(asmprate, "Audio Sample rate(616 only) 0-44.1KHz 1-48KHz,"
		 "2-32KHz (default="
		 __MODULE_STRING(DFLT_AUDIO_SMPRATE) ")");
MODULE_PARM_DESC(smode, "Audio mode 0-stereo 1-joint stereo"
		 " 2-dual channel 3-single channel (default="
		 __MODULE_STRING(DFLT_STEREO_MODE) ")");
MODULE_PARM_DESC(avsel, "Multiplexed stream 0-audio/video"
		 " 1-video only 2-audio only (default="
		 __MODULE_STRING(DFLT_AV_SELECT) ")");
MODULE_PARM_DESC(gpio, "GPIO setting (default ="
		 __MODULE_STRING(DFLT_GPIO) ")");
/* microphone settings*/
MODULE_PARM_DESC(ampgain, "Audio input gain when AGC off"
		 "(0=-3dB to 127(+60.5dB) step 0.5dB (default="
		 __MODULE_STRING(DFLT_INAMPGAIN) ")");

MODULE_PARM_DESC(agcon, "Audio AGC on/off 0-on, 1-off (default="
		 __MODULE_STRING(DFLT_AGCON) ")");
MODULE_PARM_DESC(agcsens, "Audio AGC sensitivity 0-6 (default="
		 __MODULE_STRING(DFLT_AGCSENS) ")");

MODULE_PARM_DESC(nosubid, "Fix for certain 616's that have no subvendor-id"
		 " 1-enable 0-disable (default=0)");

MODULE_AUTHOR("Sensoray Co. Inc.");
MODULE_DESCRIPTION("Driver for Sensoray's model 516 and 616"
		   " MPEG CODEC boards (Release: MODULE_REVISION)");
MODULE_SUPPORTED_DEVICE("{{Sensoray,sx16}}");

/* Get release from automatically updated release string*/
static char *GetRelease(void)
{
	/*static char AutoRelease[] = "$Release: dev_1.2 $";	???*/
	/*AutoRelease[sizeof(AutoRelease) - 2] = 0;*/
	/*return &AutoRelease[9];*/
	return "Release: " MODULE_REVISION;
}


/* Get ISA bus speed information.*/
/* This information is returned when in /proc/sx16*/
#define SPEEDBUF_SIZE ((unsigned long)4000)
#if LINUX_VERSION_CODE < KERNEL_VERSION(3, 10, 0)
static int print_ISA_speed(int board, char *buf, int len)
{
	void *tbuf;
	struct timeval tv_start, tv_stop;
	unsigned long xfertime;
	unsigned long xferrate;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(5, 0, 0 )
	struct timespec64 ts;
#endif
	/* Measure bus speed if 516.*/
	if (model_num != 616) {
		tbuf = kmalloc(SPEEDBUF_SIZE,	GFP_KERNEL | __GFP_DMA);

		if (tbuf == NULL)
			return len;

		if (CORE_acquire_mutex(hbd[board]) < 0)
			return len;

		/* Get start time.*/
#if LINUX_VERSION_CODE < KERNEL_VERSION(5, 0, 0 )
		do_gettimeofday(&tv_start);
#else
		ktime_get_real_ts64(&ts);
		tv_start.tv_sec = ts.tv_sec;
		tv_start.tv_usec = ts.tv_nsec / NSEC_PER_USEC;
#endif		

		/* Read bytes.*/
		insw(io[board] + 6, tbuf, SPEEDBUF_SIZE / 2);
		/* Get Stop time.*/
#if LINUX_VERSION_CODE < KERNEL_VERSION(5, 0, 0 )		
		do_gettimeofday(&tv_stop);
#else
		ktime_get_real_ts64(&ts);
		tv_stop.tv_sec = ts.tv_sec;
		tv_stop.tv_usec = ts.tv_nsec / NSEC_PER_USEC;
#endif		
		CORE_release_mutex(hbd[board]);

		len += sprintf(buf + len, "Maximum ISA bus transfer rate:\n");
		xfertime = tv_stop.tv_usec - tv_start.tv_usec;
		if (xfertime > 0) {
			xferrate = (SPEEDBUF_SIZE * 1000000) / xfertime;
			len += sprintf(buf + len,
				       "Read:%ld bytes/sec (%ld bits/sec)\n",
				       xferrate, xferrate * 8);
		}

		if (CORE_acquire_mutex(hbd[board]) < 0)
			return len;

		/* Get start time.*/
		do_gettimeofday(&tv_start);
		/* Read 32K bytes.*/
		outsw(io[board] + 6, tbuf, SPEEDBUF_SIZE / 2);
		/* Get Stop time.*/
		do_gettimeofday(&tv_stop);

		CORE_release_mutex(hbd[board]);
		xfertime = tv_stop.tv_usec - tv_start.tv_usec;
		if (xfertime > 0) {
			xferrate = (SPEEDBUF_SIZE * 1000000) / xfertime;
			len += sprintf(buf + len, "Write: %ld bytes/sec (%ld bits/sec)\n", xferrate, xferrate * 8);
		}
		kfree(tbuf);
	}
	return len;
}
#else
static int print_ISA_speed(int board, struct seq_file *m)
{
	void *tbuf;
	struct timeval tv_start, tv_stop;
	unsigned long xfertime;
	unsigned long xferrate;

	/* Measure bus speed if 516.*/
	if (model_num != 616) {
		tbuf = kmalloc(SPEEDBUF_SIZE,	GFP_KERNEL | __GFP_DMA);

		if (tbuf == NULL)
			return 0;

		if (CORE_acquire_mutex(hbd[board]) < 0)
			return 0;

		/* Get start time.*/
		do_gettimeofday(&tv_start);
		/* Read bytes.*/
		insw(io[board] + 6, tbuf, SPEEDBUF_SIZE / 2);
		/* Get Stop time.*/
		do_gettimeofday(&tv_stop);

		CORE_release_mutex(hbd[board]);

		seq_printf(m, "Maximum ISA bus transfer rate:\n");
		xfertime = tv_stop.tv_usec - tv_start.tv_usec;
		if (xfertime > 0) {
			xferrate = (SPEEDBUF_SIZE * 1000000) / xfertime;
			seq_printf(m, "Read:%ld bytes/sec (%ld bits/sec)\n",
				   xferrate, xferrate * 8);
		}

		if (CORE_acquire_mutex(hbd[board]) < 0)
			return 0;

		/* Get start time.*/
		do_gettimeofday(&tv_start);
		/* Read 32K bytes.*/
		outsw(io[board] + 6, tbuf, SPEEDBUF_SIZE / 2);
		/* Get Stop time.*/
		do_gettimeofday(&tv_stop);

		CORE_release_mutex(hbd[board]);
		xfertime = tv_stop.tv_usec - tv_start.tv_usec;
		if (xfertime > 0) {
			xferrate = (SPEEDBUF_SIZE * 1000000) / xfertime;
			seq_printf(m, "Write: %ld bytes/sec (%ld bits/sec)\n", xferrate, xferrate * 8);
		}
		kfree(tbuf);
	}
	return 0;
}
#endif
/* MPEG queue polling function.
   This function, which runs on a secondary kernel thread, polls*/
/* the queue level and wakes the poll method when the queue
   has reached the required level.*/
#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 5, 0)
static void poll_thread(kthread_t *kthread)
#else
static int poll_thread(void *vp)
#endif
{
	/* Setup the thread environment.*/
#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 5, 0)
	long board = kthread->board;
	HBD h_hbd = hbd[board];
	init_kthread(kthread, kthread->arg);
	/* Wait for things to settle...*/
#else
	wait_queue_head_t poll_sleep_q; /* sleep wait queue.*/
	long board = (long) vp;
	HBD h_hbd = hbd[board];
	DEFINE_WAIT(wait);
	init_waitqueue_head(&poll_sleep_q);
#endif
	/* Repeat until the driver is closed ...*/
	do {
		/* memory barrier */
		mb();
		/* Check the appropriate queue level.
		   Wake up the wait queue if (1) app is reading
		   (encoding) and there is mpeg data available,
		   or (2) app is writing (decoding)
		   and there is space to add more mpeg data.*/
		if (CORE_stream_is_encoding(h_hbd)) {
			if (CORE_get_dma_remaining(h_hbd) >= 4096 || (CORE_queue_get_any_size(h_hbd, Q_ID_MUX) > LOW_WATERMARK && CORE_poll_dma_flag(h_hbd))) {
				wake_up_interruptible(&stream_q[board]);
			}
		} else {
			if (CORE_stream_is_decoding(h_hbd)) {
				if (CORE_get_dma_remaining(h_hbd) >= 4096 || (CORE_queue_get_any_size(h_hbd, Q_ID_MUX) < HIGH_WATERMARK_VIDEO && CORE_poll_dma_flag(h_hbd)))
					wake_up_interruptible(&stream_q[board]);
			}
		}
#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 5, 0)
		interruptible_sleep_on_timeout(&kthread->queue, (10 * HZ) / 1000);
    }
    while (!kthread->terminate);

    exit_kthread(kthread);
    return;
#else
       prepare_to_wait(&poll_sleep_q, &wait, TASK_INTERRUPTIBLE);
       schedule_timeout(5*HZ);
       finish_wait(&poll_sleep_q, &wait);
    }
	while (!kthread_should_stop());

    return 0;
#endif
}

/* ioctl method - the 'misc.' function of the driver.*/

/* Helper function: map ioctl number into core's attribute identifier.*/
static int a_id(unsigned int ioctl_num)
{

	switch (ioctl_num) {
	case SX16_IOC_SET_AV_SELECT:
		return ATTR_av_select;
	case SX16_IOC_SET_PROG_NUM:
		return ATTR_program_number;
	case SX16_IOC_SET_INPUT:
		return ATTR_video_chan;
	case SX16_IOC_SET_INPUT_TYPE:
		return ATTR_video_type;
	case SX16_IOC_SET_VIDEO_SYSTEM:
		return ATTR_video_system;
	case SX16_IOC_SET_BITRATE:
		return ATTR_video_bitrate;
	case SX16_IOC_SET_M:
		return ATTR_video_m;
	case SX16_IOC_SET_N:
		return ATTR_video_n;
	case SX16_IOC_SET_VBR:
		return ATTR_video_vbr_enable;
	case SX16_IOC_SET_BRIGHTNESS:
		return ATTR_video_brightness;
	case SX16_IOC_SET_CONTRAST:
		return ATTR_video_contrast;
	case SX16_IOC_SET_SATURATION:
		return ATTR_video_saturation;
	case SX16_IOC_SET_COLORBARS:
		return ATTR_video_colorbars_enable;
	case SX16_IOC_SET_FILTER:
		return ATTR_video_filter_enable;
	case SX16_IOC_SET_AUDIO_VOLUME:
		return ATTR_audio_volume;
	case SX16_IOC_SET_AUDIO_MUTE:
		return ATTR_audio_mute;
	case SX16_IOC_SET_AUDIO_BITRATE:
		return ATTR_audio_bitrate;
	case SX16_IOC_SET_AUDIO_SMPRATE:
		return ATTR_audio_sample_rate;
	case SX16_IOC_SET_AUDIO_FORMAT:
		return ATTR_audio_format;
	case SX16_IOC_SET_AUDIO_STEREO:
		return ATTR_audio_stereo_mode;
	case SX16_IOC_SET_MPEG:
		return ATTR_mpeg;
	case SX16_IOC_SET_STREAM_TYPE:
		return ATTR_stream_type;
	case SX16_IOC_SET_AUDIO_AGCON:
		return ATTR_audio_agc_onoff;
	case SX16_IOC_SET_AUDIO_AGCSENS:
		return ATTR_audio_agc_sens;
	case SX16_IOC_SET_AUDIO_AMPGAIN:
		return ATTR_audio_inampgain;
	case SX16_IOC_SET_AUDIO_TREBLE:
		return ATTR_audio_treble;
	case SX16_IOC_SET_AUDIO_BASS:
		return ATTR_audio_bass;
	default:
		return -1;
	}
}

/* Helper function: map ioctl number into core's caption attribute identifier.*/
static int a_captn_id(unsigned int ioctl_num)
{
	switch (ioctl_num) {
	case SX16_IOC_SET_CAPTION_WINDOW:
		return CAPTN_window;
	case SX16_IOC_SET_CAPTION_XSTART:
		return CAPTN_xstart;
	case SX16_IOC_SET_CAPTION_YSTART:
		return CAPTN_ystart;
	case SX16_IOC_SET_CAPTION_SCALE:
		return CAPTN_scale;
	case SX16_IOC_SET_CAPTION_COLOR:
		return CAPTN_color;
	case SX16_IOC_SET_CAPTION_TRANSPARENCY:
		return CAPTN_transp;
	case SX16_IOC_SET_CAPTION_TEXT:
		return CAPTN_text;
	case SX16_IOC_SET_CAPTION_ON:
		return CAPTN_on;
	default:
		return -1;
	}
}

/* The public ioctl function for MPEG structure(not V4L supported here)*/
#if LINUX_VERSION_CODE >= KERNEL_VERSION_NOINODE
static long sx16_ioctl(struct file *file, unsigned int cmd, unsigned long arg)
#else
static int sx16_ioctl(struct inode *inode, struct file *file,
                      unsigned int cmd, unsigned long arg)
#endif
{
	unsigned long val = (unsigned long) arg;
	/* Board number.*/
#if LINUX_VERSION_CODE >= KERNEL_VERSION_NOINODE
#if LINUX_VERSION_CODE < KERNEL_VERSION(3, 19, 0)
	struct inode *inode = file->f_dentry->d_inode;
#else
	struct inode *inode = file_inode(file);
#endif
#endif
	int board = MINOR(inode->i_rdev);
	int retval = 0;

	dprintk(8, "IOCTL: board %d cmd %d %x(%c) %d %d\n", board,
		_IOC_DIR(cmd), _IOC_TYPE(cmd), _IOC_TYPE(cmd),
		_IOC_NR(cmd), _IOC_SIZE(cmd));

	/* Decode the ioctl command code. see sx16ioctl.h for descriptions*/
	switch (cmd) {
	case SX16_IOC_GET_BTYPE:
		retval = model_num;
		break;
	case SX16_IOC_SET_BRIGHTNESS:
	case SX16_IOC_SET_COLORBARS:
	case SX16_IOC_SET_CONTRAST:
	case SX16_IOC_SET_FILTER:
	case SX16_IOC_SET_INPUT:
	case SX16_IOC_SET_INPUT_TYPE:
	case SX16_IOC_SET_SATURATION:
	case SX16_IOC_SET_VIDEO_SYSTEM:
	case SX16_IOC_SET_AUDIO_BITRATE:
	case SX16_IOC_SET_AUDIO_SMPRATE:
	case SX16_IOC_SET_AUDIO_FORMAT:
	case SX16_IOC_SET_AV_SELECT:
	case SX16_IOC_SET_PROG_NUM:
	case SX16_IOC_SET_BITRATE:
	case SX16_IOC_SET_M:
	case SX16_IOC_SET_N:
	case SX16_IOC_SET_VBR:
	case SX16_IOC_SET_MPEG:
	case SX16_IOC_SET_STREAM_TYPE:
	case SX16_IOC_SET_AUDIO_VOLUME:
	case SX16_IOC_SET_AUDIO_MUTE:
	case SX16_IOC_SET_AUDIO_STEREO:
	case SX16_IOC_SET_AUDIO_TREBLE:
	case SX16_IOC_SET_AUDIO_BASS:
	case SX16_IOC_SET_AUDIO_AGCON:
	case SX16_IOC_SET_AUDIO_AGCSENS:
	case SX16_IOC_SET_AUDIO_AMPGAIN:
		/* Program attribute setting.*/
		CORE_attr_set(hbd[board], a_id(cmd), val);
		break;
	case SX16_IOC_GET_SETTING:
		/* Return current attribute setting.*/
		retval = CORE_attr_get(hbd[board], val);
		break;
	case SX16_IOC_SET_CAPTION_TEXT:
		/* Set caption text to overlay*/
		/* Verify that we can directly access the user text buffer;
		   abort with error indication*/
		/* if direct access is not possible.*/
		if (!ACCESS_OK(VERIFY_READ, (void *)val, 80)) {
			retval = -EFAULT;
			break;
		}
		/* fall through */
	case SX16_IOC_SET_CAPTION_WINDOW:
	case SX16_IOC_SET_CAPTION_XSTART:
	case SX16_IOC_SET_CAPTION_YSTART:
	case SX16_IOC_SET_CAPTION_SCALE:
	case SX16_IOC_SET_CAPTION_COLOR:
	case SX16_IOC_SET_CAPTION_TRANSPARENCY:
	case SX16_IOC_SET_CAPTION_ON:
		CORE_caption_set(hbd[board], a_captn_id(cmd), val);
		break;
	case SX16_IOC_GET_CAPTION_SETTINGS:
		retval = CORE_caption_get(hbd[board], val);
		break;
	/* STREAM CONTROL ------------------------------------*/
	case SX16_IOC_RW_STOP:
		/* Stop read/write.*/
		CORE_stream_stop(hbd[board]);
		break;
	case SX16_IOC_RW_PAUSE:
		/* Pause read/write.*/
		CORE_stream_pause(hbd[board]);
		break;
	case SX16_IOC_RW_RESUME:
		/* Resume read/write after pause.*/
		CORE_stream_resume(hbd[board]);
		break;
	/* DIGITAL IO ------------------------------------*/
	case SX16_IOC_GET_GPIO:
		/*
		 * Read state of GPIO pins.
		 * Bit is set to 1 when input is low.
		 */
		retval = CORE_gpio_read(hbd[board]);
		break;
	case SX16_IOC_SET_GPIO:
		/*
		 * Set state of GPIO pins.
		 * Bit is set to 1 to drive output low.
		 */
		CORE_gpio_write(hbd[board], val);
		break;
	case SX16_IOC_GA_LOOPBACK:
		/*
		 * Enable/disable video stream loopback through gate array
		 *   (bypasses SM).
		 */
		CORE_stream_bypass(hbd[board], val);
		break;
	case SX16_IOC_RESET:
		/* Reset the board.*/
		CORE_board_reset(hbd[board]);
		dbgprint("reset ioctl\n");
		break;
	case SX16_IOC_GET_Q_SIZE:
		/* Return level of selected queue.*/
		retval = CORE_queue_get_any_size(hbd[board], val);
		break;
	case SX16_IOC_RD_SENSORAY:
		/*
		 * Transfer encoded mpeg data
		 * (in Sensoray packetized format) to user buffer.
		 */
		retval = read_sensoray(hbd[board], (char *)val,
				       SENSORAY_PAYLOAD_SIZE);
		break;
	case SX16_IOC_RD_SENSORAY_TS:
		/*
		 * Transfer encoded mpeg data (in Sensoray packetized format)
		 *  to user buffer.
		 */
		retval = read_sensoray(hbd[board], (char *)val,
				       SENSORAY_PAYLOAD_SIZE_TS);
		break;
#ifdef SX16_BUILD_SRAY
	case SX16_IOC_WR_SENSORAY_NET:
		/*
		 * Transfer encoded mpeg data (in Sensoray packetized format)
		 * from user buffer.
		 */
		if (CORE_attr_get(hbd[board], ATTR_stream_type) == STREAMTYPE_VES)
			retval = write_sensoray(hbd[board], (char *)val, 1);
		else if (CORE_attr_get(hbd[board], ATTR_stream_type) == STREAMTYPE_MUX)
			retval = write_sensoray_mux(hbd[board], (char *)val, 1);
		else
			retval = -EINVAL;
		break;
	case SX16_IOC_WR_SENSORAY_FILE:
		/*
		 * Transfer encoded mpeg data (in Sensoray packetized format)
		 * from user buffer.
		 */
		if (CORE_attr_get(hbd[board], ATTR_stream_type)
		    == STREAMTYPE_VES)
			retval = write_sensoray(hbd[board], (char *)val, 0);
		else if (CORE_attr_get(hbd[board], ATTR_stream_type) == STREAMTYPE_MUX)
			retval = write_sensoray_mux(hbd[board], (char *)val, 0);
		else
			retval = -EINVAL;
		break;
	case SX16_IOC_RD_RTP:
		/* Transfer encoded mpeg data (in RTP format) to user buffer.*/
		/* check user space region .*/
		if (!ACCESS_OK(VERIFY_WRITE , (void *)val, 32768))
			retval = -EFAULT;
		else
			retval = CORE_read_rtp(hbd[board], (char *) val);
		break;
	case SX16_IOC_RTP_PLUS:
		/* If val!=NULL raw data sent to *val along with RTP;
		   if val==NULL - not sent*/
		if (!ACCESS_OK(VERIFY_WRITE, (void *)val, 2*32768))
			retval =  -1;
		else
			retval = CORE_rtp_plus(hbd[board], (char *) val);
		break;
	/* RTCP/SSCP*/
	case SX16_IOC_GET_RTCP_SR:
	case SX16_IOC_GET_RTCP_RR:
	case SX16_IOC_PUT_RTCP_RR:
	case SX16_IOC_PUT_RTCP_SR:
	case SX16_IOC_GET_SSCP_SR:
	case SX16_IOC_GET_SSCP_RR:
	case SX16_IOC_PUT_SSCP_RR:
	case SX16_IOC_PUT_SSCP_SR:
	case SX16_IOC_GET_RTCPA_SR:
		retval = RTCP_SSCP_Report(hbd[board], (void *) val, cmd);
#endif
		break;
	case SX16_IOC_GET_INPUT_STATUS:
		retval = CORE_in_status_get(hbd[board]);
		break;
	case SX16_IOC_WAIT_VIDEO:
		/* Wait for video sync (616 only).--------------*/
		video_sleepq(hbd[board], val);
		break;
	case SX16_IOC_SET_ENCODER_OUT:
		/* Video output shut off control. -------------------*/
		CORE_encoder_out_ctrl(hbd[board], (int) val);
		break;
	case SX16_IOC_DECODER_READY:
		retval = CORE_stream_is_decoding(hbd[board]);
		break;
	case SX16_IOC_ENCODER_READY:
		retval = CORE_stream_is_encoding(hbd[board]);
		break;
	case SX16_IOC_START_DECODER:
		if (!CORE_stream_is_decoding(hbd[board])) {
			if (val)
				start_decoder(board, 1); /* non-blocking*/
			else
				CORE_start_decode(hbd[board]);/* blocks*/
			retval = 0;
		} else
			retval = -1;
		break;
	case SX16_IOC_START_ENCODER:
		if (!CORE_stream_is_encoding(hbd[board])) {
			CORE_start_encode(hbd[board]);
			retval = 0;
		} else
			retval = -1;
		break;
	default:
		dprintk(2, "sx16::UNSUPPORTED IOCTL %d\n", cmd);
		return -ENOIOCTLCMD;
	}
	return retval;
}


#if defined(SX16_V4L_BUILD)

static int vidioc_querycap(struct file *file, void *priv,
			   struct v4l2_capability *cap)
{
	struct sx16_v4l *v4l = NULL;
	v4l = video_drvdata(file);	
	strlcpy(cap->driver, "sx16v", sizeof(cap->driver));
	strlcpy(cap->card, "sx16v", sizeof(cap->card));
	cap->version = SX16_VERSION;
#if LINUX_VERSION_CODE > KERNEL_VERSION(3, 3, 0)
	cap->capabilities = V4L2_CAP_VIDEO_CAPTURE | V4L2_CAP_STREAMING | V4L2_CAP_READWRITE | V4L2_CAP_DEVICE_CAPS;
#else	
	cap->capabilities = V4L2_CAP_VIDEO_CAPTURE | V4L2_CAP_STREAMING | V4L2_CAP_READWRITE;
#endif
	snprintf(cap->bus_info, sizeof(cap->bus_info), "PCI:%s", pci_name(v4l->pci));
	return 0;
}

static int vidioc_enum_fmt_cap(struct file *file, void *priv,
			       struct v4l2_fmtdesc *f)
{
	int index = 0;
	if (f)
		index = f->index;

	if (index >= ARRAY_SIZE(formats))
		return -EINVAL;

	//memset(f, 0, sizeof(struct v4l2_fmtdesc));
	dprintk(4, "name %s\n", formats[index].name);
	strlcpy(f->description, formats[index].name, sizeof(f->description));
	f->pixelformat = formats[index].fourcc;
	f->type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
	f->index = index;
	f->flags = 0;
	return 0;
}

static int vidioc_g_fmt_cap(struct file *file, void *priv,
			    struct v4l2_format *f)
{
	struct sx16_fh *fh = priv;
	if (fh == NULL)
		return -EFAULT;

	if (fh->fmt == NULL)
		return -EFAULT;
	f->fmt.pix.width = fh->width;
	f->fmt.pix.height = fh->height;
	f->fmt.pix.field = fh->cap.field;
	f->fmt.pix.pixelformat = fh->fmt->fourcc;
	f->fmt.pix.bytesperline = f->fmt.pix.width * (fh->fmt->depth >> 3);
	f->fmt.pix.sizeimage = f->fmt.pix.height * f->fmt.pix.bytesperline;
	f->fmt.pix.colorspace = V4L2_COLORSPACE_SMPTE170M;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(4, 2, 0)
	f->fmt.pix.ycbcr_enc = V4L2_YCBCR_ENC_DEFAULT;
	f->fmt.pix.quantization = V4L2_QUANTIZATION_DEFAULT;
	f->fmt.pix.xfer_func = V4L2_XFER_FUNC_DEFAULT;
#endif	
	return 0;
}



static const struct sx16_fmt *format_by_fourcc(int fourcc)
{
	unsigned int i;

	for (i = 0; i < ARRAY_SIZE(formats); i++) {
		if (-1 == formats[i].fourcc)
			continue;
		if (formats[i].fourcc == fourcc)
			return formats + i;
	}
	return NULL;
}


static int vidioc_try_fmt_cap(struct file *file, void *priv,
			      struct v4l2_format *f)
{
	const struct sx16_fmt *fmt;
	enum v4l2_field field;
	int  b_any_field = 0;
	//struct sx16_fh *fh = priv;
	//int board = fh->board;
	int maxh, maxw;
	int is_ntsc;
	struct sx16_v4l *v4l = NULL;
	v4l = video_drvdata(file);	

	dprintk(1, "%s\n", __func__);

	is_ntsc = (v4l->is_ntsc); //(current_norm & V4L2_STD_NTSC) ? 1 : 0;
	fmt = format_by_fourcc(f->fmt.pix.pixelformat);

	if (fmt == NULL) {
		
		return -EINVAL;
	}

	field = f->fmt.pix.field;
	if (field == V4L2_FIELD_ANY)
		b_any_field = 1;

	dprintk(4, "try format %d \n", is_ntsc);
	/* supports 3 sizes. see sx16drv.h */
	dprintk(50, "width test %d, height %d\n",
		f->fmt.pix.width, f->fmt.pix.height);


	maxw  = is_ntsc ? SYS_XACTIVE_MAX_NTSC : SYS_XACTIVE_MAX_PAL;
	maxh  = is_ntsc ? SYS_YACTIVE_MAX_NTSC : SYS_YACTIVE_MAX_PAL;


	if (f->fmt.pix.width > MAX_WIDTH)
		f->fmt.pix.width = MAX_WIDTH;
	else if (f->fmt.pix.width < 56)
		f->fmt.pix.width = 56;

	if (f->fmt.pix.height > maxh)
		f->fmt.pix.height = maxh;
	else if (f->fmt.pix.height < 32)
		f->fmt.pix.height = 32;

	/* enforce 1/2 size limits if TOP or BOTTOM */
	if ((field == V4L2_FIELD_TOP) || (field == V4L2_FIELD_BOTTOM)) {
		if (f->fmt.pix.width > maxw/2)
			f->fmt.pix.width = maxw/2;
		if (f->fmt.pix.height > maxh/2)
			f->fmt.pix.height = maxh/2;
	}

	if (V4L2_FIELD_ANY == field) {
		field = (f->fmt.pix.height > maxh/2)
			? V4L2_FIELD_INTERLACED
			: V4L2_FIELD_TOP;
	} else {
		if ((f->fmt.pix.height > maxh/2) &&
		    (field != V4L2_FIELD_INTERLACED))
			return -EINVAL;
		if ((f->fmt.pix.height <= maxh/2) &&
		    (field == V4L2_FIELD_INTERLACED))
			return -EINVAL;
	}

	dprintk(4, "field interlaced %d\n", field == V4L2_FIELD_INTERLACED);
	f->fmt.pix.field = field;
	f->fmt.pix.bytesperline = f->fmt.pix.width * (fmt->depth >> 3);
	f->fmt.pix.sizeimage = f->fmt.pix.height * f->fmt.pix.bytesperline;
	dprintk(4, "width %d height %d field %d \n", f->fmt.pix.width,
		f->fmt.pix.height, f->fmt.pix.field);

	f->fmt.pix.colorspace = V4L2_COLORSPACE_SMPTE170M;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(4, 2, 0)
	f->fmt.pix.ycbcr_enc = V4L2_YCBCR_ENC_DEFAULT;
	f->fmt.pix.quantization = V4L2_QUANTIZATION_DEFAULT;	
	f->fmt.pix.xfer_func = V4L2_XFER_FUNC_DEFAULT;
#endif
	return 0;
}

static int vidioc_s_fmt_cap(struct file *file, void *priv,
			    struct v4l2_format *f)
{
	struct sx16_fh *fh = priv;
	const struct sx16_fmt *fmt;
	int ret;

	ret = vidioc_try_fmt_cap(file, fh, f);

	if (ret < 0)
		return (ret);

	fmt = format_by_fourcc(f->fmt.pix.pixelformat);
	if (fmt == NULL)
		return -EINVAL;

	if (res_locked(fh->v4l, RESOURCE_VIDEO)) {
		dprintk(1, "can't change format after started\n");
		return -EBUSY;
	}

	fh->fmt = fmt;
	fh->width = f->fmt.pix.width;
	fh->height = f->fmt.pix.height;
	fh->cap.field = f->fmt.pix.field;
	fh->type = f->type;
	CORE_capture_init(fh->v4l->hbd,
			  fh->height,
			  fh->width,
			  fh->fmt->depth,
			  fh->fmt->fourcc);
	return 0;
}

static int vidioc_reqbufs(struct file *file, void *priv,
			  struct v4l2_requestbuffers *p)
{
	int rc;
	struct sx16_fh *fh = priv;
	rc = videobuf_reqbufs(&fh->cap, p);
	return rc;
}

static int vidioc_querybuf(struct file *file, void *priv, struct v4l2_buffer *p)
{
	int rc;
	struct sx16_fh *fh = priv;
	rc = videobuf_querybuf(&fh->cap, p);
	return rc;
}

static int vidioc_qbuf(struct file *file, void *priv, struct v4l2_buffer *p)
{
	int rc;
	struct sx16_fh *fh = priv;
	rc = videobuf_qbuf(&fh->cap, p);
	return rc;
}

static int vidioc_dqbuf(struct file *file, void *priv, struct v4l2_buffer *p)
{
	int rc;
	struct sx16_fh *fh = priv;
	rc = videobuf_dqbuf(&fh->cap, p, file->f_flags & O_NONBLOCK);
	return rc;
}


#ifdef CONFIG_VIDEO_V4L1_COMPAT
static int vidioc_cgmbuf(struct file *file, void *priv, struct video_mbuf *mbuf)
{
	struct sx16_fh *fh = priv;
	struct videobuf_queue *q;
	struct v4l2_requestbuffers req;
	int rc, i;
	int count = 8;
	q = &fh->cap;
	memset(&req, 0, sizeof(req));
	req.type = q->type;
	req.count = count;
	req.memory = V4L2_MEMORY_MMAP;
	rc = videobuf_reqbufs(q, &req);
	if (rc < 0)
		return rc;

	mbuf->frames = req.count;
	mbuf->size   = 0;
	for (i = 0; i < mbuf->frames; i++) {
		mbuf->offsets[i] = q->bufs[i]->boff;
		mbuf->size += q->bufs[i]->bsize;
	}
	return 0;
}
#endif

static int vidioc_streamon(struct file *file, void *priv, enum v4l2_buf_type i)
{
	struct sx16_fh *fh = priv;
	int retval;
	dprintk(1, "%s\n", __func__);
	if (fh == NULL)
		return -EFAULT;

	if (irq[fh->board] <= 0) {
		printk(KERN_ERR "must enable interrupts to use V4L\n");
		return -EFAULT;
	}

	if (!res_get(fh, RESOURCE_VIDEO)) {
		dprintk(1, "stream already on\n");
		return -EBUSY;
	}
	/* save our context */
	fh->v4l->video_fh = fh;
	retval = videobuf_streamon(&fh->cap);
	dprintk(4, "stream on %d width %d height %d\n", retval, fh->width,
		fh->height);
	return retval;
}

static int vidioc_streamoff(struct file *file, void *priv,
			    enum v4l2_buf_type i)
{
	struct sx16_fh *fh = priv;
	int retval;
	if (fh == NULL)
		return -EFAULT;

	if (!res_locked(fh->v4l, RESOURCE_VIDEO)) {
		dprintk(0, "vidioc_streamoff: not streaming\n");
		return -EBUSY;
	}

	CORE_capture_stop(fh->v4l->hbd);

	retval = videobuf_streamoff(&fh->cap);
	del_timer(&fh->v4l->video_q.timeout);

	dprintk(2, "streamoff %d\n", retval);
	res_free(fh, RESOURCE_VIDEO);
	return retval;
}

#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 20)
static int vidioc_s_std_old(struct file *file, void *priv, v4l2_std_id i)
{
	/* videodev in the kernel is buggy for this for 2.6.18 and 2.6.19 only
	   return 0 and then our ioctl will call the normal one 
	*/
	return 0;

}
#endif


#if LINUX_VERSION_CODE >= KERNEL_VERSION(3, 10, 0) || \
    LINUX_VERSION_CODE <= KERNEL_VERSION(2, 6, 16) || \
    LINUX_VERSION_CODE == KERNEL_VERSION(2, 6, 18)
static int vidioc_s_std(struct file *file, void *priv, v4l2_std_id _i)
{
	v4l2_std_id *i = &_i;
#else
static int vidioc_s_std(struct file *file, void *priv, v4l2_std_id *i)
{
#endif
	int std;
	struct sx16_fh *fh = priv;
	v4l2_std_id norm = *i;
	struct sx16_v4l *v4l = NULL;
	v4l = video_drvdata(file);
	
	/* convert from v4l2 to internal standard*/
	dprintk(2, "%s %d \n", __func__, (int) norm);
	if (res_locked(fh->v4l, RESOURCE_VIDEO)) {
		dprintk(0, "vidioc_s_std: queue busy\n");
		return -EBUSY;
	}

	std = v4l2_to_sray_std(norm);
	/* check valid*/
	if (std == -1) {
		printk(KERN_INFO "%s invalid standard\n", __func__);
		return -EINVAL;
	}
	/* set the standard*/
	CORE_attr_set(fh->v4l->hbd, a_id(SX16_IOC_SET_VIDEO_SYSTEM), std);
	v4l->is_ntsc = (std == SYS_NTSC);
	//printk("is ntsc s_std %d\n", v4l->is_ntsc);
	return 0;
}


static int vidioc_g_std(struct file *file, void *priv, v4l2_std_id *i)
{
	struct sx16_v4l *v4l = NULL;
	v4l = video_drvdata(file);
	*i = v4l->is_ntsc ? V4L2_STD_NTSC : V4L2_STD_PAL;
	//printk("%s, is_ntsc %d\n", __func__, v4l->is_ntsc);
	dprintk(1, "%s %d (0x%x)\n", __func__, (int)*i, (int)*i);
	// Note: this code below may not be necessary because
	// if you are using vidioc_g_std, you shouldn't be using
	// current_norm.
#ifdef HAVE_CURRENT_NORM
	dev->vdev.current_norm = *i;
#endif
	return 0;
}
 
static int vidioc_enum_input(struct file *file, void *priv,
			     struct v4l2_input *inp)
{
	int i = inp->index;
	if (i < 0 || i >= NUM_CAP_INPUTS)
		return -EINVAL;

	switch (i) {
	case 0:
		memset(inp, 0, sizeof(struct v4l2_input));
		strlcpy(inp->name, "C-Video (upper BNC connector)",
			 sizeof(inp->name));
		inp->type = V4L2_INPUT_TYPE_CAMERA;
		inp->std =  V4L2_STD_NTSC | V4L2_STD_PAL;
#if LINUX_VERSION_CODE > KERNEL_VERSION(2, 6, 32)
		inp->capabilities = V4L2_IN_CAP_STD;
#endif
		inp->index = 0;
		break;
	case 1:
		memset(inp, 0, sizeof(struct v4l2_input));
		strlcpy(inp->name, "S-Video (middle connector)",
			 sizeof(inp->name));
		inp->type = V4L2_INPUT_TYPE_CAMERA;
		inp->std =  V4L2_STD_NTSC | V4L2_STD_PAL;
#if LINUX_VERSION_CODE > KERNEL_VERSION(2, 6, 32)
		inp->capabilities = V4L2_IN_CAP_STD;
#endif		
		inp->index = 1;
		break;
	default:
		break;
	}
	return 0;

}


static int vidioc_g_input(struct file *file, void *priv, unsigned int *i)
{
	struct sx16_fh *fh = priv;

	dprintk(2, "%s: %d \n", __func__, *i);
	*i = CORE_attr_get(fh->v4l->hbd, a_id(SX16_IOC_SET_INPUT_TYPE));
	return 0;
}

static int vidioc_s_input(struct file *file, void *priv, unsigned int i)
{
	struct sx16_fh *fh = priv;
	if (i > 1) {
		return -EINVAL;
	}
	dprintk(2, "set input %p %d\n", priv , i);
	if (!fh)
	  return -ENODEV;
	if (!fh->v4l)
	  return -ENODEV;
	if (!fh->v4l->hbd)
	  return -ENODEV;		
	CORE_attr_set(fh->v4l->hbd, a_id(SX16_IOC_SET_INPUT_TYPE), i);
	return 0;
}


/* --- controls ---------------------------------------------- */
static int vidioc_queryctrl(struct file *file, void *priv,
			    struct v4l2_queryctrl *qc)
{
	int i;

	for (i = 0; i < ARRAY_SIZE(video_ctrls); i++)
		if (qc->id && qc->id == video_ctrls[i].id) {
			memcpy(qc, &(video_ctrls[i]),
			       sizeof(*qc));
			return (0);
		}

	return -EINVAL;
}

static int vidioc_g_ctrl(struct file *file, void *priv,
			 struct v4l2_control *c)
{
	struct sx16_fh	*fh  = priv;
	if (!c) {
		return -EINVAL;
	}

	switch (c->id) {
	case V4L2_CID_BRIGHTNESS:
		c->value = CORE_attr_get(fh->v4l->hbd, ATTR_video_brightness);
		break;
	case V4L2_CID_CONTRAST:
		c->value = CORE_attr_get(fh->v4l->hbd, ATTR_video_contrast);
		break;
	case V4L2_CID_SATURATION:
		c->value = CORE_attr_get(fh->v4l->hbd, ATTR_video_saturation);
		break;
	default:
		return -EINVAL;
	}
	return 0;
}

static int vidioc_s_ctrl(struct file *file, void *priv,
			 struct v4l2_control *c)
{
	struct sx16_fh *fh = priv;

	switch (c->id) {
	case V4L2_CID_BRIGHTNESS:
		CORE_attr_set(fh->v4l->hbd, ATTR_video_brightness, c->value);
		break;
	case V4L2_CID_CONTRAST:
		CORE_attr_set(fh->v4l->hbd, ATTR_video_contrast, c->value);
		break;
	case V4L2_CID_SATURATION:
		CORE_attr_set(fh->v4l->hbd, ATTR_video_saturation, c->value);
		break;
	default:
		return -EINVAL;
	}
	return 0;
}

#if LINUX_VERSION_CODE < KERNEL_VERSION_NOINODE
static int video_ioctl(struct inode *inode, struct file *file, unsigned int cmd, unsigned long arg)
#else
static long video_ioctl(struct file *file, unsigned int cmd, unsigned long arg)
#endif
{

	int retval = 0;
	struct sx16_fh	*fh  = file->private_data;
	struct sx16_v4l *v4l = NULL;
	/* safety checks*/
	if (fh == NULL)
		return -EFAULT;

	if (fh)	{
		v4l = fh->v4l;
		if (v4l == NULL)
			return -EFAULT;
	}
	/* Decode the ioctl command code.*/
	dprintk(2, "ioctl %d\n", _IOC_NR(cmd));
#if LINUX_VERSION_CODE >= KERNEL_VERSION_NOINODE
	retval = video_ioctl2(file, cmd, arg);
#else
	retval = video_ioctl2(inode, file, cmd, arg);
#endif
#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 20)
/* kernel versions from 2.6.18 and 2.6.19 have a bug in
   the common videodev VIDIOC_S_STD routine.  see vidioc_s_std_old
   This bug was fixed in Linux kernel 2.6.20.  Leave this patch for
   2.6.18 and 2.6.19.
*/
	if (cmd == VIDIOC_S_STD) {
		v4l2_std_id id = *(v4l2_std_id *)arg;
		dprintk(2, "s_std 2.6.16 kernel %d\n", id);
		retval = vidioc_s_std(file, fh, &id);
		return retval;

	}
#endif

	if (retval == -EINVAL) {
		int rc2;
#if LINUX_VERSION_CODE >= KERNEL_VERSION_NOINODE		
		rc2 = sx16_ioctl(file, cmd, arg);
#else
		rc2 = sx16_ioctl(inode, file, cmd, arg);
#endif
		if (rc2 != -ENOIOCTLCMD)
			retval = rc2;

	}
	return retval;
}

#endif /* else NO_SX16_V4L*/

/*
 * READ METHOD - Attempts to read "nbytes" bytes of data from the core's
 * MPEG buffer.  If this is  the first read and the encoder is not already
 * running, the encoder is initialized and started.
 * Return value:
 *	0	: Blocking enabled and we timed out while waiting for data.
 *	-1	: Blocking disabled and insufficient data is available,
 *                or can't copy to user buffer.
 *	N	: Successful, N bytes were read.
 *                N may be less than nbytes if nbytes exceeds the max
 *               legal transfer size
 *               (in which case N will be the max legal transfer size).
 */

static ssize_t sx16_read(struct file *file, char *buffer, size_t nbytes,
			 loff_t *ppos)
{
#ifdef DBG
	if (dbgdata.bytes + nbytes > DBGDATA_SIZE)
		nbytes = DBGDATA_SIZE - dbgdata.bytes;
	if (nbytes) {
		memcpy(buffer, (void *) (dbgdata.data + dbgdata.bytes), nbytes);
		dbgdata.bytes += nbytes;
	}
	return nbytes;
#else
	int		rc;				/* Returned byte count.*/
#if LINUX_VERSION_CODE < KERNEL_VERSION(3, 19, 0)
	int		board			= MINOR(file->f_dentry->d_inode->i_rdev);
#else
	struct inode *inode = file_inode(file);
	int board = MINOR(inode->i_rdev);
#endif
	int		no_block		= (file->f_flags & O_NONBLOCK) != 0;

	/*
	 * Verify that we can directly access the user buffer;
	 * abort with error indication if direct access is not possible.
	 */
	if (!ACCESS_OK(VERIFY_WRITE, (void *)buffer, nbytes)) {
		dprintk(4, "access not ok, read \n");
		return -EFAULT;
	}

	rc = sx16_read_core(board, buffer, nbytes, no_block);
	return rc;
#endif
}



#if defined(SX16_V4L_BUILD)
#if 0
static ssize_t
video_read(struct file *file, char __user *data, size_t count, loff_t *ppos)
{
	struct sx16_fh *fh = file->private_data;

	switch (fh->type) {
	case V4L2_BUF_TYPE_VIDEO_CAPTURE:
		if (res_locked(fh->v4l, RESOURCE_VIDEO))
			return -EBUSY;
		return videobuf_read_one(&fh->cap,
					 data, count, ppos,
					 file->f_flags & O_NONBLOCK);
		break;
	default:
		return 0;
	}
}
#endif
#endif



/*
 * Special READ method (implemented as ioctl):
 * formats mpeg data into "Sensoray packets" as it is written out to
 * user space buffer.
 * Attempts to read NUM_SENSORAY_PACKETS worth of data from the core's
 * MPEG buffer.	 If this is the first read and the encoder is not already
 * running, the encoder is initialized and started.  Unlike the standard Read
 * method, this is ALWAYS a blocking call.
 * Return value:
 *	0	: Timed out while waiting for data.
 *	-1	: Can't copy to user buffer.
 *	N	: Successful, N packets were read.
 */
static ssize_t read_sensoray(HBD board, char *buffer, int payloadsize)
{
	int i; /* Packet transfer loop counter.*/
	/* Packet sequence number.  not init for pseudo-randomness.*/
	static unsigned short sequence_num;
	/* board identifier */
	int board_num = -1;
	int rc;
	unsigned long *pulong;
	unsigned short *pushort;
	i = 0;

	if ((payloadsize != SENSORAY_PAYLOAD_SIZE) && (payloadsize != SENSORAY_PAYLOAD_SIZE_TS)) {
		payloadsize = SENSORAY_PAYLOAD_SIZE;
	}

	board_num = handle_to_boardnum(board);

	if (board_num == -1)
		return -EFAULT;

	/*
	 * Verify that we can directly access the user buffer;
	 * abort with error indication if direct access is not possible.
	 */
	if (!ACCESS_OK(VERIFY_WRITE, (void *)buffer,
		       sizeof(int) + NUM_SENSORAY_PACKETS * (SENSORAY_HEADER_SIZE + payloadsize)))
		return -EFAULT;

	/* Start the encoder if it is not already running.*/
	if (!CORE_stream_is_encoding(board)) {
		/* re-init the sequence number.*/
		/* otherwise the decoder will have serious problems if encoder restarts*/
		sequence_num = (unsigned short) SX16MOD_time_now();
		CORE_start_encode(board);
	}

	/*
	 * Copy packet count to packet group header.
	 * Change to __put_user() for portability.
	 */
	pulong = (unsigned long *) buffer;
	*pulong = NUM_SENSORAY_PACKETS;
	buffer += sizeof(unsigned long);

	/* For each packet ...*/
	for (i = 0; i < NUM_SENSORAY_PACKETS; i++) {
		/*
		 * Copy packet length and sequence number to packet header,
		 * bump sequence number.
		 * Change to __put_user() for portability.
		 */
		pulong = (unsigned long *) buffer;
		*pulong = SENSORAY_HEADER_SIZE + payloadsize;
		buffer += sizeof(unsigned long);
		pushort = (unsigned short *) buffer;
		*pushort = sequence_num++;
		buffer += sizeof(unsigned short);
		rc = sx16_read_core(board_num, buffer, payloadsize, 0);
		/*
		 * Copy packet payload (mpeg data) to user space.
		 * Abort with error if copy failed.
		 */
		if (rc != payloadsize)
			return -EAGAIN;
		else
			buffer += payloadsize;
	}

	PacketCountRx[board_num] += i;
	/*
	 * set the next 4 bytes of buffer to 0
	 * (indicates no audio present to server)
	 */

	*(int *)buffer = 0;
	/* Indicate number of mpeg bytes transferred.*/
	return NUM_SENSORAY_PACKETS;
}

#ifdef SX16_BUILD_SRAY
static ssize_t write_sensoray(HBD board, char *buffer, int skip_data)
{
	int	Packets = *(int *) buffer;	/*number of packets in buffer*/
	int board_num = handle_to_boardnum(board);

	/*
	 * Verify that we can directly access the user buffer;
	 * abort with error indication if direct access is not possible.
	 */
	if (!ACCESS_OK(VERIFY_READ, (void *)buffer,
		       sizeof(int) + Packets * (SENSORAY_HEADER_SIZE + SENSORAY_PAYLOAD_SIZE)))
		return -EFAULT;
	/*
	 * function below will reset the sensoray write state machine
	 * if x16 is not decoding
	 */
	while (start_decoder(board_num, skip_data)) {
		if (skip_data)
			return Packets;
	}
	return SX16CORE_write_sensoray(board, buffer, skip_data);
}

static ssize_t write_sensoray_mux(HBD board, char *buffer, int skip_data)
{
	int	Packets = *(int *) buffer;	/*number of packets in buffer*/
	int board_num = handle_to_boardnum(board);
	/*
	 * Verify that we can directly access the user buffer;
	 * abort with error indication if direct access is not possible.
	 */

	if (!ACCESS_OK(VERIFY_READ, (void *)buffer,
		       sizeof(int) + Packets * (SENSORAY_HEADER_SIZE + SENSORAY_PAYLOAD_SIZE)))
		return -EFAULT;
	/*
	 * function below will reset the sensoray write_mux state machine
	 * if x16 is not decoding
	 */
	while (start_decoder(board_num, skip_data)) {
		if (skip_data)
			return Packets;
	}
	return SX16CORE_write_sensoray_mux(board, buffer, skip_data);
}
#endif
/*
 *  WRITE METHOD - Attempts to write "nbytes" bytes of data to the core's
 * MPEG buffer.  If this is the first write and the decoder is not already
 * running, the decoder is initialized and started.
 * Return value:
 * 0: Blocking enabled and we timed out while waiting for buffer space.
 * -1: Blocking disabled and insufficient space is available,
 * or can't copy to user buffer.
 * N: Successful, N bytes were written.
 * N may be less than nbytes if nbytes exceeds the max
 * legal transfer size (in which case N will be the max legal transfer size).
 */
static ssize_t sx16_write(struct file *file, const char *buffer,
			  size_t nbytes, loff_t *ppos)
{
	int	rc;	/* Returned byte count.*/
#if LINUX_VERSION_CODE < KERNEL_VERSION(3, 19, 0)
	int	board = MINOR(file->f_dentry->d_inode->i_rdev);
#else
	struct inode *inode = file_inode(file);
	int board = MINOR(inode->i_rdev);
#endif

	int	no_block = (file->f_flags & O_NONBLOCK) != 0;
	HBD	h_hbd = hbd[board];
	if (!ACCESS_OK(VERIFY_READ, (void *) buffer, nbytes)) {
		dprintk(4, "efault");
		return -EFAULT;
	}
	/* Start the decoder if it is not already running.*/
	if (!CORE_stream_is_decoding(h_hbd)) {
		CORE_start_decode(h_hbd);
		write_count[board] = 0;
	}
	rc = sx16_write_core(board, buffer, nbytes, no_block);
	return rc;
}


#if defined(SX16_V4L_BUILD)
#if 0
static ssize_t video_write(struct file *file, const char *buffer, size_t nbytes, loff_t *ppos)
{
	int board = MINOR(file->f_dentry->d_inode->i_rdev);
	printk(KERN_INFO "V4L write not supported, board num%d\n", board);
	return -EINVAL;
}
#endif
#endif


/*
 * OPEN METHOD - check that the minor device is valid. If not, return
 * ENODEV.  Increment usage counter (many processes may use this device
 * simultaneously), allocate buffer.
 */


static int sx16_open(struct inode *inode, struct file *file)
{
	int board = MINOR(inode->i_rdev);
	int res;
	dprintk(2, "%s\n", __func__);
	if (down_interruptible(&sx16_devices_lock))
		return -ERESTARTSYS;
	dprintk(2, "OPEN: board %d\n", board);
	/* Check validity of minor (board) number.*/
	if (((unsigned int)board > num_boards) || !board_installed[board]) {
		printk(KERN_ERR "%s  invalid minor device number or driver"
		       " not loaded yet %d %d %d\n", __func__,
               board, num_boards, board_installed[board]);
		up(&sx16_devices_lock);
		return -ENODEV;
	}

	file->private_data = NULL;
	if (open_count_mpeg[board] > 0) {
		open_count_mpeg[board]++;
		up(&sx16_devices_lock);
		return 0;
	}
	/* Init the board and take it to its default state.*/
	res = CORE_board_open(hbd[board]);
	if ((res == FALSE) && (open_count_v4l[board] == 0)) {
		printk(KERN_INFO "failed to open board core resources"
		       " %d %d %d!\n",
		       res, open_count_v4l[board], board);
		up(&sx16_devices_lock);
		return -ENODEV;
	}

#ifdef DBG
	dbgdata.bytes = 0;
#endif
	/* Start the poll thread, which manages the MPEG queue level.*/
#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 5, 0)
	poll_thread_tcb[board].arg = poll_thread_name[board];
	poll_thread_tcb[board].board = board;
	start_kthread(poll_thread, &poll_thread_tcb[board]);
	poll_thread_exists[board] = TRUE; /* Set pollling thread flag.*/
#else
	poll_thread_tcb[board] = kthread_run(poll_thread, (void *)(long)board, poll_thread_name[board]);
	if (poll_thread_tcb[board] < 0) {
		poll_thread_tcb[board] = NULL;
		printk(KERN_ERR "Unable to start kernel thread for X16.\n");
		up(&sx16_devices_lock);
		return -ENODEV;
	}
#endif

	/* One more app is using this module.*/
#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 5, 0)
	MOD_INC_USE_COUNT;
#endif
	open_count_mpeg[board]++;
	up(&sx16_devices_lock);
	dprintk(4, "sx16_open %d\n", open_count_mpeg[board]);
	return 0;
}


#if defined(SX16_V4L_BUILD)

/*
 * OPEN METHOD - check that the minor device is valid. If not, return
 * ENODEV.  Increment usage counter (many processes may use this device
 * simultaneously), allocate buffer.
 */
#if LINUX_VERSION_CODE >= KERNEL_VERSION_NOINODE
static int video_open(struct file *file)
#else
static int video_open(struct inode *inode, struct file *file)  
#endif
{
	struct sx16_v4l *v4l = NULL;
	struct sx16_fh *fh = NULL;
	int res;
	int board;
	v4l = video_drvdata(file);
	if (v4l == NULL) {
		printk("board not setup yet\n");
		return -EAGAIN;
	}
	board = v4l->board;
	dprintk(4, "[%d]%s\n", board, __func__);

	
	dprintk(4, "open v4l %d\n", board);
	if (down_interruptible(&sx16_devices_lock))
		return -ERESTARTSYS;

	if (board_installed[board] != TRUE) {
		printk("board not installed yet\n");
		up(&sx16_devices_lock);
		return -EAGAIN;
	}
	/* Check validity of minor (board) number.*/
	if (((unsigned int)board > num_boards) || !board_installed[board]) {
		printk(KERN_ERR "Error: invalid minor device number.\n");
		up(&sx16_devices_lock);
		return -ENODEV;
	}
	/* if 616, open per context file date*/
	/* allocate the frame header*/
	if (model_num != 616) {
		printk(KERN_ERR "sx16: not 616\n");
		up(&sx16_devices_lock);
		/* V4L only for 616*/
		return -ENODEV;
	}

	if (open_count_v4l[board] != 0) {
		/* One more app is using this module.*/
		printk(KERN_ERR "only one open board handle at a time\n");
		up(&sx16_devices_lock);
		return 0;
	}
	fh = kzalloc(sizeof(*fh), GFP_KERNEL);
	if (NULL == fh) {
		printk(KERN_ERR "sx16: out of memory!\n");
		up(&sx16_devices_lock);
		return -ENOMEM;
	}
	/* assign to private data*/
	file->private_data = fh;
	
	*fh = v4l->init;
	fh->board = board;
	fh->v4l = v4l;
	fh->type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
	fh->width = 704;
	fh->height = 480;
	fh->fmt = &formats[0];
	fh->cap.field = V4L2_FIELD_INTERLACED;
#ifdef VIDIOC_G_PRIORITY
	v4l2_prio_open(&v4l->prio, &fh->prio);
#endif

#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 23)
	videobuf_queue_init(&fh->cap, &sx16_video_qops,
			    s616_dev[board], &v4l->slock,
			    V4L2_BUF_TYPE_VIDEO_CAPTURE,
			    V4L2_FIELD_INTERLACED,
			    sizeof(struct sx16_buffer), fh);
#elif LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 26)
	videobuf_queue_pci_init(&fh->cap, &sx16_video_qops,
				s616_dev[board], &v4l->slock,
				V4L2_BUF_TYPE_VIDEO_CAPTURE,
				V4L2_FIELD_INTERLACED,
				sizeof(struct sx16_buffer), fh);

#elif LINUX_VERSION_CODE < KERNEL_VERSION(3, 0, 0)
/* changed again as of 2.6.26-2.6.27 */
	videobuf_queue_sg_init(&fh->cap, &sx16_video_qops,
				&s616_dev[board]->dev, &v4l->slock,
				V4L2_BUF_TYPE_VIDEO_CAPTURE,
				V4L2_FIELD_INTERLACED,
				sizeof(struct sx16_buffer), fh);
#else
	videobuf_queue_sg_init(&fh->cap, &sx16_video_qops,
			       &s616_dev[board]->dev, &v4l->slock,
			       V4L2_BUF_TYPE_VIDEO_CAPTURE,
			       V4L2_FIELD_INTERLACED,
			       sizeof(struct sx16_buffer), fh,
			       NULL);	
#endif

/* Init the board and take it to its default state.*/
/* board_open will fail if any instance has the board open.*/
/* so don't need to check open_count_mpeg here*/
	res = CORE_board_open(hbd[board]);
	if ((res == FALSE) && (open_count_mpeg[board] == 0)) {
		printk(KERN_ERR "failed to open board core resources %d %d %d %d!\n",
		       res, open_count_v4l[board],
		       open_count_mpeg[board],
		       board);
		kfree(fh);
		kfree(v4l);
		up(&sx16_devices_lock);
		return -ENODEV;
	}
#ifdef DBG
	dbgdata.bytes = 0;
#endif
	/* One more app is using this module.*/
	open_count_v4l[board]++;
	up(&sx16_devices_lock);
	return 0;
}
#endif

/*
 * CLOSE METHOD - release the device and decrement the usage counters.
 * release all board releases(open_count_mpeg == 0, open_count_v4l == 0)
 */

static int SX16MOD_release(int board, struct file *file)
{
#if defined(SX16_V4L_BUILD)
	struct sx16_fh	*fh  = file->private_data;
	struct sx16_v4l *v4l = NULL;
	dprintk(4, "mod release %d\n", board);
	if (fh)
		v4l = fh->v4l;
#endif

#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 5, 0)
	if (poll_thread_exists[board]) {
		stop_kthread(&poll_thread_tcb[board]);
		poll_thread_exists[board] = FALSE;
	}
	if (start_thread_flag[board]) {
		stop_kthread(&start_thread_tcb[board]);
		start_thread_flag[board] = 0;
	}
#else
	if (poll_thread_tcb[board]) {
		kthread_stop(poll_thread_tcb[board]);
		poll_thread_tcb[board] = NULL;
	}
	if (start_thread_tcb[board]) {
		kthread_stop(start_thread_tcb[board]);
		start_thread_tcb[board] = NULL;
	}
#endif
	/* Return board to its default state.*/
	CORE_board_close(hbd[board]);
	dprintk(2, "sx16: close board %d. hbd: %p\n", board, hbd[board]);
	return 0;
}


static int sx16_release(struct inode *inode, struct file *file)
{
	int board = MINOR(inode->i_rdev);
	if (down_interruptible(&sx16_devices_lock)) {
		dprintk(4, "%s interrupted\n", __func__);
		return -ERESTARTSYS;
	}
	file->private_data = NULL;
	if (open_count_mpeg[board] > 1)	{
		open_count_mpeg[board]--;
#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 5, 0)
		MOD_DEC_USE_COUNT;
#endif
		up(&sx16_devices_lock);
		return 0;
	}
#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 5, 0)
	MOD_DEC_USE_COUNT;
#endif
	open_count_mpeg[board] = 0;
	/* release all boards*/
	if (open_count_v4l[board] == 0) {
		SX16MOD_release(board, file);
	}
	up(&sx16_devices_lock);
	return 0;
}

#if defined(SX16_V4L_BUILD)
/* V4L release function*/
#if LINUX_VERSION_CODE >= KERNEL_VERSION_NOINODE	
static int video_release(struct file *file)
#else
static int video_release(struct inode *inode, struct file *file)
#endif
{
	struct sx16_fh	*fh  = file->private_data;
	struct sx16_v4l *v4l = NULL;
	int board;
	v4l = video_drvdata(file);
	board = v4l->board;

	if (!fh) {
		printk(KERN_INFO "%s: no filehandle fh\n", __func__);
		return -ENODEV;
	}
	dprintk(4, "vdeo release board %d v4l %p\n", board, v4l);
	if (down_interruptible(&sx16_devices_lock)) {
		printk(KERN_INFO "%s interrupted\n", __func__);
		return -ERESTARTSYS;
	}
	dprintk(4, "%s : v4l open count %d sx16 open count %d\n",
		__func__,
		open_count_v4l[board],
		open_count_mpeg[board]);

	if (open_count_v4l[board] > 0) {
		if (res_check(fh, RESOURCE_VIDEO)) {
			dprintk(4, "stopping video_release\n");
			CORE_capture_stop(fh->v4l->hbd);
			videobuf_streamoff(&fh->cap);
			del_timer(&fh->v4l->video_q.timeout);
			res_free(fh, RESOURCE_VIDEO);
			v4l->video_fh = NULL;
		}
		CORE_capture_close(fh->v4l->hbd);
		/* release all boards*/
		if (open_count_mpeg[board] == 0) {
			SX16MOD_release(board, file);
		}
		if (fh)
			kfree(fh);

		file->private_data = NULL;
		open_count_v4l[board]--;
	} else {
		dprintk(4, "already released V4L\n");
	}
	up(&sx16_devices_lock);
	return 0;
}


/* for MMAP*/
static int video_mmap(struct file *file, struct vm_area_struct *vma)
{
	struct sx16_fh *fh = file->private_data;
	return videobuf_mmap_mapper(&fh->cap, vma);
}

#endif
/*
 * POLL METHOD - polls the device for readiness.
 * for MPEG device
 */
unsigned int sx16_poll(struct file *file, poll_table *wait)
{
	/* Get the board handle.*/
#if LINUX_VERSION_CODE < KERNEL_VERSION(3, 19, 0)
	int board = MINOR(file->f_dentry->d_inode->i_rdev);
#else
	struct inode *inode = file_inode(file);
	int board = MINOR(inode->i_rdev);
#endif
	HBD h_hbd = hbd[board];

	/*
	 * Return with all flags set if the encoder and decoder
	 *   are not running to prevent indefinite blocking.
	 */

	if (!(CORE_stream_is_encoding(h_hbd)
	      || CORE_stream_is_decoding(h_hbd)))
		return POLLIN | POLLRDNORM | POLLOUT | POLLWRNORM;


	/*
	 * Return appropriate flags if the MPEG queue is ready to
	 * send/receive data.
	 */
	if (CORE_poll_dma_flag(hbd[board])) {
		if (CORE_stream_is_encoding(h_hbd)) {
			if ((CORE_get_dma_remaining(h_hbd) >= 4096)
			    || (CORE_queue_get_any_size(h_hbd, Q_ID_MUX)
				> LOW_WATERMARK))
				return POLLIN | POLLRDNORM;
		} else {
			if ((CORE_get_dma_remaining(h_hbd) >= 4096)
			    || (CORE_queue_get_any_size(h_hbd, Q_ID_MUX)
			    < HIGH_WATERMARK_VIDEO))
				return POLLOUT | POLLWRNORM;
		}
	}
	/* Add this board's list of blocked processes to the poll queue.*/
	poll_wait(file, &stream_q[board], wait);
	/*
	 * Return appropriate flags
	 * if the MPEG queue is ready to send/receive data.
	 */

	if (CORE_poll_dma_flag(h_hbd)) {
		if (CORE_stream_is_encoding(h_hbd)) {
			if ((CORE_get_dma_remaining(h_hbd) >= 4096)
			    || (CORE_queue_get_any_size(h_hbd, Q_ID_MUX) > LOW_WATERMARK))
				return POLLIN | POLLRDNORM;
		} else {
			if ((CORE_get_dma_remaining(h_hbd) >= 4096)
			    || (CORE_queue_get_any_size(h_hbd, Q_ID_MUX) < HIGH_WATERMARK_VIDEO))
				return POLLOUT | POLLWRNORM;
		}
	}
	/* Indicate Not Ready.*/
	return 0;
}


/* POLL METHOD - polls the device for readiness.*/
/* for V4l device*/
#if defined(SX16_V4L_BUILD)

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 25)
#define MUTEX_LOCK_VB(fh) mutex_lock(&fh->cap.vb_lock)
#else
#define MUTEX_LOCK_VB(fh) mutex_lock(&fh->cap.lock)
#endif

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 25)
#define MUTEX_UNLOCK_VB(fh) mutex_unlock(&fh->cap.vb_lock)
#else
#define MUTEX_UNLOCK_VB(fh) mutex_unlock(&fh->cap.lock)
#endif

static unsigned int video_poll(struct file *file,
			       struct poll_table_struct *wait)
{
	struct sx16_fh *fh = file->private_data;
	struct videobuf_buffer *buf = NULL;

	if (fh == NULL)
		return -ENODEV;

	if (res_check(fh, RESOURCE_VIDEO)) {
		if (!list_empty(&fh->cap.stream))
			buf = list_entry(fh->cap.stream.next,
					 struct videobuf_buffer, stream);
	} else {
		MUTEX_LOCK_VB(fh);

		if (UNSET == fh->cap.read_off) {
			/* need to capture a new frame */
			if (res_locked(fh->v4l, RESOURCE_VIDEO)) {
				MUTEX_UNLOCK_VB(fh);
				return POLLERR;
			}
			if (0 != fh->cap.ops->buf_prepare(&fh->cap,
							  fh->cap.read_buf,
							  fh->cap.field)) {
				MUTEX_UNLOCK_VB(fh);
				return POLLERR;
			}
			fh->cap.ops->buf_queue(&fh->cap, fh->cap.read_buf);
			fh->cap.read_off = 0;
		}
		MUTEX_UNLOCK_VB(fh);
		buf = fh->cap.read_buf;
	}

	if (!buf)
		return POLLERR;

	poll_wait(file, &buf->done, wait);
	if (buf->state == VIDEOBUF_DONE ||
	    buf->state == VIDEOBUF_ERROR)
		return POLLIN|POLLRDNORM;
	return 0;
}

#endif



/* Map attribute values to text equivalents.*/
static char txt_muxformat[][40] = {"svcd", "vcd", "dvd", "cdda", "ves", "aes"
				   , "tran" };
static char txt_arate	 [][40] = {"", "32", "48", "56", "64", "80", "96",
				   "112", "128", "160", "192", "224", "256",
				   "320", "384" };
static char txt_vbr	 [][40] = {"const", "var" };
static char txt_vsystem	 [][40] = {"", "pal", "ntsc" };
static char txt_intype	 [][40] = {"comp", "svid" };
static char txt_sysformat[][40] = {"svcd", "vcd", "other" };
static char txt_avsel	 [][40] = {"audio+video", "video-only", "audio-only" };
static char txt_asmprate [][40] = {"44", "48", "32" };
static char txt_mpeg	 [][40] = {"mpeg2:D1_704", "mpeg2:2/3_D1",
				   "mpeg2:1/2_D1", "mpeg1:SIF", "mpeg1:QSIF",
				   "mpeg2:D1_720", "mpeg2:D1_640" };

/*
 * This function generates an entry in the file system's /proc directory.
 * The information provided here is useful for conveying status
 * and diagnostics info to the outside world.
 * Returns length of output string.
 */
#ifndef CONFIG_PROC_FS
static int sx16_read_procmem(char *buf, char **start, off_t offset,
			     int count, int *eof, void *data)
{
	printk(KERN_INFO "proc fs not configured\n");
	return 0;
}
#else
#if LINUX_VERSION_CODE < KERNEL_VERSION(3, 10, 0)
int sx16_read_procmem(char *buf, char **start, off_t offset, int count,
		      int *eof, void *data)
{
	int i, j;
	/* Milliseconds elapsed since stream started.*/
	unsigned long ms_elapsed;
	/* Formatted, elapsed time since stream started.*/
	unsigned long h, m, s, t, x;
	int len = 0; /* Char count in generated string.*/
	int ucode_vers;	/* Microcode version.*/
	int is_encoding; /* Board is encoding.*/
	int is_decoding; /* Board is decoding.*/
	unsigned long KB_xferred; /* Number of mpeg KB transferred.*/
	unsigned long bit_rate; /* Bit rate in Kb/sec.*/

	/* Determine max number of characters that our generated string
	   may have.  To prevent kernel*/
	/* buffer overflow, we must ensure that our string does
	   not exceed length (PAGE_SIZE - count).*/
	int limit = PAGE_SIZE - offset - 120;

	/* Generate report heading.*/
	len += sprintf(buf + len, "Sensoray X16 drvr, %d bd(s) det:\n",
		       num_boards);

	/* Generate a report for each board in the system ...*/
	for (i = 0; (i < num_boards) && (len <= limit); i++) {
		HBD h_hbd = hbd[i];
		if (board_installed[i]) {
			write_param_t writeParam;
			CORE_get_write_params(h_hbd, &writeParam);
			/* Fetch board status;*/
			ucode_vers = CORE_microcode_version(h_hbd);
			is_encoding = CORE_stream_is_encoding(h_hbd);
			is_decoding = CORE_stream_is_decoding(h_hbd);
			ms_elapsed = (is_encoding || is_decoding) ?
				CORE_get_elapsed_time(h_hbd) : 0;

			t = ms_elapsed / 1000;
			h = t / 3600;
			x = t - h * 3600;
			m = x / 60;
			s = x - m * 60;
			KB_xferred = 8 * CORE_get_stat_KB(h_hbd);
			/* fix for divide by zero error on startup. */
			if (t != 0)
				bit_rate = (is_encoding || is_decoding) ?
					(KB_xferred / t) : 0;
			else
				bit_rate = 0;
			/* Board heading.*/
			len += sprintf(buf + len,
				       " bd%i: mdl=%d mod=", i, model_num);

			/* Stream mode (microcode version),
			   run time and measured bit rate.*/
			if (is_encoding)
				len += sprintf(buf + len, "enc");
			else if (is_decoding)
				len += sprintf(buf + len, "dec");
			else
				len += sprintf(buf + len, "stby");

			if (is_encoding || is_decoding) {
				len += sprintf(buf + len,
					       " fw=v.%X.%2.2X",
					       ucode_vers >> 8,
					       ucode_vers & 0xFF);

				len += sprintf(buf + len, " run=%2.2ld:%2.2ld:%2.2ld", h, m, s);
				len += sprintf(buf + len, " mbr=%ld kbps", bit_rate);
			}

			/* I/O addresses.*/
			len += sprintf(buf + len, " pa=0x%8.8lX va=0x%8.8lX irq=%2.2d", io[i], (unsigned long)log_base[i], irq[i]);
			/* Report current settings.*/
			len += sprintf(buf + len, " ch=%ld", CORE_attr_get(h_hbd, ATTR_video_chan));
			len += sprintf(buf + len, " typ=%s",  txt_intype[CORE_attr_get(h_hbd, ATTR_video_type)]);
			len += sprintf(buf + len, " vsys=%s", txt_vsystem[CORE_attr_get(h_hbd, ATTR_video_system)]);
			len += sprintf(buf + len, " sfmt=%s", txt_sysformat[CORE_attr_get(h_hbd, ATTR_video_system_format)]);
			len += sprintf(buf + len, " mfmt=%s", txt_muxformat[CORE_attr_get(h_hbd, ATTR_mux_format)]);
			len += sprintf(buf + len, " vbr=%ld", CORE_attr_get(h_hbd, ATTR_video_bitrate));
			len += sprintf(buf + len, "/%s", txt_vbr[CORE_attr_get(h_hbd, ATTR_video_vbr_enable)]);
			len += sprintf(buf + len, " abr=%s", txt_arate[CORE_attr_get(h_hbd, ATTR_audio_bitrate)]);
			len += sprintf(buf + len, " asr=%s", txt_asmprate[CORE_attr_get(h_hbd, ATTR_audio_sample_rate)]);
			len += sprintf(buf + len, " sel=%s", txt_avsel[CORE_attr_get(h_hbd, ATTR_av_select)]);
			len += sprintf(buf + len, " mpg=%s", txt_mpeg			[CORE_attr_get(h_hbd, ATTR_mpeg)]);
			len += sprintf(buf + len, " qv=%ld", CORE_queue_get_any_size(h_hbd, Q_ID_VID));
			len += sprintf(buf + len, " qa=%ld", CORE_queue_get_any_size(h_hbd, Q_ID_AUD));
			len += sprintf(buf + len, " qm=%ld", CORE_queue_get_any_size(h_hbd, Q_ID_MUX));
			/* GPIO read*/
			len += sprintf(buf + len, " gpio=%x", CORE_gpio_read(h_hbd));
			/* double buffer counter */
			len += sprintf(buf + len, " dblbuf=%d", writeParam.dblBufferCount);
			/* lost packets*/
			len += sprintf(buf + len, " lp=%d", writeParam.LostPackets);
			/* extra stream back to back*/
			len += sprintf(buf + len, " upl=%ld,%ld", writeParam.last_upload_time[0], writeParam.last_upload_time[1]);

			len += sprintf(buf + len, " syn=%ld,%ld", writeParam.last_sync_time[0], writeParam.last_sync_time[1]);
			len += sprintf(buf + len, " pkt=%ld,%ld", writeParam.last_packetlost_time[0], writeParam.last_packetlost_time[1]);
			len += sprintf(buf + len, " db1=%ld,%ld", writeParam.last_dbgtime1[0], writeParam.last_dbgtime1[1]);
			len += sprintf(buf + len, " db2=%ld,%ld", writeParam.last_dbgtime2[0], writeParam.last_dbgtime2[1]);
			for (j = 0; j < 11; j++) {
				len += sprintf(buf + len, " dbg[%d]=%d", j, writeParam.last_error[j]);
			}

			/* Prepare for next board.*/
			len += sprintf(buf + len, "\n");
		}
	}

	/* Insert a linefeed.*/
	len += sprintf(buf + len, "\n");

	/* Print the ISA speed for the first installed ISA board*/
	for (i = 0; (i < num_boards) && (len <= limit); i++) {
		if (board_installed[i]) {
			int retlen = print_ISA_speed(i, buf, len);
			if (retlen > len) {
				len = retlen;
				break;
			}
		}
	}

	/* Insert a linefeed.*/
	len += sprintf(buf + len, "\n");

	/* Indicate there is nothing more after the above generated chunk of text.*/
	*eof = 1;

	/* Return char count of the generated string.*/
	return len;
}
#else
static int sx16_read_show(struct seq_file *m, void *v);
static int sx16_proc_open(struct inode *inode, struct file *file) 
{
	return single_open(file, sx16_read_show, PDE_DATA(inode));

}
static const struct file_operations sx16_procmem_fops = {
	.open = sx16_proc_open,
	.read = seq_read,
	.llseek = seq_lseek,
	.release = seq_release
};

static int sx16_read_show(struct seq_file *m, void *v)
{
	int i, j;
	/* Milliseconds elapsed since stream started.*/
	unsigned long ms_elapsed;
	/* Formatted, elapsed time since stream started.*/
	unsigned long h, m1, s, t, x;
	int ucode_vers;	/* Microcode version.*/
	int is_encoding; /* Board is encoding.*/
	int is_decoding; /* Board is decoding.*/
	unsigned long KB_xferred; /* Number of mpeg KB transferred.*/
	unsigned long bit_rate; /* Bit rate in Kb/sec.*/
	/* Generate report heading.*/
	seq_printf(m, "Sensoray X16 drvr, %d bd(s) det:\n",
		   num_boards);

	/* Generate a report for each board in the system ...*/
	for (i = 0; i < num_boards; i++) {
		HBD h_hbd = hbd[i];
		if (board_installed[i]) {
			write_param_t writeParam;
			CORE_get_write_params(h_hbd, &writeParam);
			/* Fetch board status;*/
			ucode_vers = CORE_microcode_version(h_hbd);
			is_encoding = CORE_stream_is_encoding(h_hbd);
			is_decoding = CORE_stream_is_decoding(h_hbd);
			ms_elapsed = (is_encoding || is_decoding) ?
				CORE_get_elapsed_time(h_hbd) : 0;

			t = ms_elapsed / 1000;
			h = t / 3600;
			x = t - h * 3600;
			m1 = x / 60;
			s = x - m1 * 60;
			KB_xferred = 8 * CORE_get_stat_KB(h_hbd);
			/* fix for divide by zero error on startup. */
			if (t != 0)
				bit_rate = (is_encoding || is_decoding) ?
					(KB_xferred / t) : 0;
			else
				bit_rate = 0;
			/* Board heading.*/
			seq_printf(m, " bd%i: mdl=%d mod=", i, model_num);

			/* Stream mode (microcode version),
			   run time and measured bit rate.*/
			if (is_encoding)
				seq_printf(m, "enc");
			else if (is_decoding)
				seq_printf(m, "dec");
			else
				seq_printf(m, "stby");

			if (is_encoding || is_decoding) {
				seq_printf(m, " fw=v.%X.%2.2X",
					   ucode_vers >> 8,
					   ucode_vers & 0xFF);
				seq_printf(m, " run=%2.2ld:%2.2ld:%2.2ld", h, m1, s);
				seq_printf(m, " mbr=%ld kbps", bit_rate);
			}
			/* I/O addresses.*/
			seq_printf(m, " pa=0x%8.8lX va=0x%8.8lX irq=%2.2d", io[i], (unsigned long)log_base[i], irq[i]);
			/* Report current settings.*/
			seq_printf(m, " ch=%ld", CORE_attr_get(h_hbd, ATTR_video_chan));
			seq_printf(m, " typ=%s",  txt_intype[CORE_attr_get(h_hbd, ATTR_video_type)]);
			seq_printf(m, " vsys=%s", txt_vsystem[CORE_attr_get(h_hbd, ATTR_video_system)]);
			seq_printf(m, " sfmt=%s", txt_sysformat[CORE_attr_get(h_hbd, ATTR_video_system_format)]);
			seq_printf(m, " mfmt=%s", txt_muxformat[CORE_attr_get(h_hbd, ATTR_mux_format)]);
			seq_printf(m, " vbr=%ld", CORE_attr_get(h_hbd, ATTR_video_bitrate));
			seq_printf(m, "/%s", txt_vbr[CORE_attr_get(h_hbd, ATTR_video_vbr_enable)]);
			seq_printf(m, " abr=%s", txt_arate[CORE_attr_get(h_hbd, ATTR_audio_bitrate)]);
			seq_printf(m, " asr=%s", txt_asmprate[CORE_attr_get(h_hbd, ATTR_audio_sample_rate)]);
			seq_printf(m, " sel=%s", txt_avsel[CORE_attr_get(h_hbd, ATTR_av_select)]);
			seq_printf(m, " mpg=%s", txt_mpeg			[CORE_attr_get(h_hbd, ATTR_mpeg)]);
			seq_printf(m, " qv=%ld", CORE_queue_get_any_size(h_hbd, Q_ID_VID));
			seq_printf(m, " qa=%ld", CORE_queue_get_any_size(h_hbd, Q_ID_AUD));
			seq_printf(m, " qm=%ld", CORE_queue_get_any_size(h_hbd, Q_ID_MUX));
			/* GPIO read*/
			seq_printf(m, " gpio=%x", CORE_gpio_read(h_hbd));
			/* double buffer counter */
			seq_printf(m, " dblbuf=%d", writeParam.dblBufferCount);
			/* lost packets*/
			seq_printf(m, " lp=%d", writeParam.LostPackets);
			/* extra stream back to back*/
			seq_printf(m, " upl=%ld,%ld", writeParam.last_upload_time[0], writeParam.last_upload_time[1]);

			seq_printf(m, " syn=%ld,%ld", writeParam.last_sync_time[0], writeParam.last_sync_time[1]);
			seq_printf(m, " pkt=%ld,%ld", writeParam.last_packetlost_time[0], writeParam.last_packetlost_time[1]);
			seq_printf(m, " db1=%ld,%ld", writeParam.last_dbgtime1[0], writeParam.last_dbgtime1[1]);
			seq_printf(m, " db2=%ld,%ld", writeParam.last_dbgtime2[0], writeParam.last_dbgtime2[1]);
			for (j = 0; j < 11; j++) {
				seq_printf(m, " dbg[%d]=%d", j, writeParam.last_error[j]);
			}

			/* Prepare for next board.*/
			seq_printf(m, "\n");
		}
	}

	/* Insert a linefeed.*/
	seq_printf(m, "\n");

	/* Print the ISA speed for the first installed ISA board*/
	for (i = 0; i < num_boards; i++) {
		if (board_installed[i]) {
			print_ISA_speed(i, m);
		}
	}
	/* Insert a linefeed.*/
	seq_printf(m, "\n");
	/* Return char count of the generated string.*/
	return 0;
}
#endif
#endif
/* The access methods are placed in a table of valid file operations.
   The MPEG port of the driver available on both the 516 and 616 supports
   open, release (close), read, write and ioctl.
*/

#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 5, 0)
struct file_operations sx16_fops = {
read:		sx16_read,
write:		sx16_write,
poll:		sx16_poll,
ioctl:		sx16_ioctl,
open:		sx16_open,
release:	sx16_release
};
#else
struct file_operations sx16_fops = {
	.owner		= THIS_MODULE,
	.read		= sx16_read,
	.write		= sx16_write,
	.poll		= sx16_poll,
#if LINUX_VERSION_CODE >= KERNEL_VERSION_NOINODE
	.unlocked_ioctl	= sx16_ioctl,
#else
	.ioctl		= sx16_ioctl,
#endif

	.open		= sx16_open,
	.release	= sx16_release
};
#endif




/* Release resources used by a board.*/

static void release_resources_516(int i)
{
	release_region(io[i], REGION_SIZE_516);
}

static void release_resources_616(int i)
{
	if (log_base[i]) {
		iounmap((void *)log_base[i]);
		log_base[i] = NULL;
	}
	if (dmabuf[i]) {
		kfree(dmabuf[i]);
		dmabuf[i] = NULL;
	}

	if (io[i]) {
		release_mem_region(io[i], REGION_SIZE_616);
		io[i] = 0;
	}
	printk(KERN_INFO "disabling resources PCI\n");
	if (s616_dev[i]) {
		pci_disable_device(s616_dev[i]);
	}
	s616_dev[i] = NULL;
}

static void release_resources(int i)
{
	/* Release the board's resources.*/
	switch (model_num) {
	case 516:
		release_resources_516(i);
		break;
	case 616:
		release_resources_616(i);
		break;
	}
}

/* Allocate resources used by a board.	Returns INSTALL_PASS or INSTALL_FAIL.*/

static int attach_board_516(int i)
{
	/* Open access to board's I/O space
	   and Abort if requested I/O space is not available.*/
	if (request_region(io[i], REGION_SIZE_516, "sx16") == NULL) {
		printk(KERN_INFO "\n516(%d) address space is already occupied: 0x%4.4lX.\n",
		       i, io[i]);
		return INSTALL_FAIL;
	}

	/* Get the logical base I/O address to be used for accessing the GA registers.*/
	log_base[i] = (void *)io[i];

	/* Abort if board signature can't be detected.*/
	if (CORE_board_is_516(io[i]) == 0) {
		printk(KERN_INFO "516(%d,0x%4.4lX) not detected.\n", i, io[i]);
		release_region(io[i], REGION_SIZE_516);
		return INSTALL_FAIL;
	}
	/* If irq line is less than zero, set it to zero.*/
	/* This disables furth irq use.*/
	if (irq[i] < 0)
		irq[i] = 0;
	/* Indicate success.*/
	return INSTALL_PASS;
}

static int attach_board_616(int i)
{
	unsigned char irq_line;

	/*  Enable device.  This is the preferred way to*/
	/* get IRQ line.  Otherwise, some PCI quirks will cause invalid IRQ to*/
	/* be mapped or due to ACPI.  see quirks.c and apci*/
	/* in linux source code.*/

	if (pci_enable_device(s616_dev[i])) {
		printk(KERN_ERR "\n616. Could not enable device.  Failing install\n");
		return INSTALL_FAIL;
	}

	/* enable bus-mastering */
	pci_set_master(s616_dev[i]);

	/* Board cannot DMA to addresses above 32bit. */
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 5, 0)
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 39)
    pci_set_dma_mask(s616_dev[i], DMA_BIT_MASK(32));
#else
    pci_set_dma_mask(s616_dev[i], DMA_32BIT_MASK);
#endif
#endif

	/* Get the physical base address of the SAA7146 application register set.*/
	/* These registers are accessed through PCI region 0(BAR 0).*/
	io[i] = pci_resource_start(s616_dev[i], 0);

#if LINUX_VERSION_CODE < KERNEL_VERSION(4, 0, 0)
	/* Abort if the requested memory space is not available.*/
	if (check_mem_region(io[i], REGION_SIZE_616)) {
		printk(KERN_ERR "\n616(%d) address space at 0x%8.8lX is already attached.\n", i, io[i]);

		return INSTALL_FAIL;
	}
#endif

	printk(KERN_DEBUG "sx16: [1]IRQ %d \n", s616_dev[i]->irq);

	/* Open access to board's memory space.*/
	if (!request_mem_region(io[i], REGION_SIZE_616, "sx16")) {
		printk(KERN_ERR "request mem region failed\n");
#ifndef TEST_PCI_616
		return INSTALL_FAIL;
#else
		return INSTALL_PASS;
#endif
	}
	/* Get the logical base memory address to be used for accessing the SAA7146 registers.*/
	log_base[i] = ioremap(io[i], REGION_SIZE_616);
	/* To use PCI interrupts irq must be set to zero (default) on the insmod command line.*/
	printk(KERN_INFO "sx16: irq value is %d\n", irq[i]);
	if (irq[i] == 0) {
		/*  get PCI irq from device structure*/
		irq_line = s616_dev[i]->irq;
		irq[i] = irq_line;
		printk(KERN_DEBUG "irq value is now %d\n", irq[i]);
#if 0
		/* commented out 8/10/2006*/
		/* Get IRQ line that system setup for 616.*/
		if (pci_read_config_byte(s616_dev[i], PCI_INTERRUPT_LINE, &irq_line)) {
			printk(KERN_ERR "\n616(%d) interrupt line failed. Using polling method.\n", i);
			irq[i] = 0;
		} else {
			irq[i] = irq_line;
		}
#endif
	} else
		irq[i] = 0;

	/* Indicate success.*/
	return INSTALL_PASS;
}

static int attach_board(int i)
{
	/* Attempt to attach to the board.*/
	switch (model_num) {
	case 516:
		return attach_board_516(i);
	case 616:
		return attach_board_616(i);
	default:
		return INSTALL_FAIL;
	}
}

/* Interrupt handler.*/

#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 5, 0)
void interrupt_handler(int irq, void *dev_id, struct pt_regs *regs)
{
	int board = *((int *)dev_id);
	int sts;
	/* Read interrupt status and reset all interrupts.*/
	sts = CORE_reset_interrupt(hbd[board]);
	/* Check status. */
	if (sts & INTERRUPT_STREAM)
		wake_up_interruptible(&stream_q[board]);	/* Wake up stream waiting processes.*/
	if (sts & INTERRUPT_VIDEO) {

		wake_up_interruptible(&video_q[board]);	/* Wake up video waiting processes.*/
#if defined(SX16_V4L_BUILD)
		if (G_v4l[board]) {
			video_irq_done(G_v4l[board], sts);
		}
#endif
	}

}
#else
#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 19)
irqreturn_t interrupt_handler(int irq, void *dev_id, struct pt_regs *regs)
#else
irqreturn_t interrupt_handler(int irq, void *dev_id)
#endif
{
	int board = *((int *)dev_id);
	int handled = 0;
	int sts;

	/* Read interrupt status and reset all interrupts.*/
	sts = CORE_reset_interrupt(hbd[board]);

	/* Check status.*/
	if (sts & INTERRUPT_STREAM) {
		handled = 1;
		wake_up_interruptible(&stream_q[board]); /* Wake up stream waiting processes.*/
	}
	if (sts & INTERRUPT_VIDEO) {
		handled = 1;
		wake_up_interruptible(&video_q[board]);	/* Wake up video waiting processes.*/
#if defined(SX16_V4L_BUILD)
		if (G_v4l[board]) {
			video_irq_done(G_v4l[board], sts);
		}
#endif
	}
	return IRQ_RETVAL(handled);
}
#endif


/* Detect and install a x16 board.  Returns INSTALL_PASS or INSTALL_FAIL.*/

static int init_board(int i)
{
	int rtnval;

	rtnval = attach_board(i);
	dmabuf[i] = 0;

	/* Allocate the 616 dma buffer.*/
	if (rtnval == INSTALL_PASS && model_num == 616)	{
		dmabuf[i] = kmalloc(CHUNK_SIZE * 2,  GFP_KERNEL | __GFP_DMA);
		if (dmabuf[i] == NULL) {
			printk(KERN_ERR "x16(%d) dma buffer allocation failed.\n", i);
			release_resources(i);
			rtnval = INSTALL_FAIL;
		}
	}
	printk(KERN_DEBUG "sx16: rtvval %d model num %d\n", rtnval, model_num);
	/* If we successfully attached to the board ...*/
	if (rtnval == INSTALL_PASS) {
		/* Attempt to init the driver core.  If init failed ...*/
		printk(KERN_DEBUG "sx16: %d base[%d] = %lx\n", (int) __LINE__, i, (long)log_base[i]);
		hbd[i] = CORE_board_init(model_num, (unsigned long)log_base[i], irq[i], (void *)dmabuf[i],
					 (unsigned long) virt_to_bus(dmabuf[i]));
		if (hbd[i] == NULL) {
			printk(KERN_ERR "x16(%d) will not initialize.\n", i);
			release_resources(i);
			rtnval = INSTALL_FAIL;
		} else {
			HBD h_hbd = hbd[i];

			/*
			 * Declare to the core all default settings that are
			 * to apply whenever the board is opened
			 *
			 * Program the settings that can be entered during insmod.
			 */
			CORE_attr_set(h_hbd, ATTR_audio_format, afmt[i]);	/* Audio format.*/
			CORE_attr_set(h_hbd, ATTR_audio_sample_rate, asmprate[i]);/* Audio sample rate.*/
			CORE_attr_set(h_hbd, ATTR_audio_stereo_mode, smode[i]);	/* Stereo mode.*/
			CORE_attr_set(h_hbd, ATTR_audio_bass, bass[i]);	/* Bass level.*/
			CORE_attr_set(h_hbd, ATTR_audio_treble,	treb[i]);	/* Treble level.*/
			CORE_attr_set(h_hbd, ATTR_audio_mute, mute[i]);	/* Mute enable.*/
			CORE_attr_set(h_hbd, ATTR_audio_volume, vol [i]); /* Volume level.*/
			CORE_attr_set(h_hbd, ATTR_video_system, system[i]); /* Video system (NTSC/PAL).*/
			CORE_attr_set(h_hbd, ATTR_video_type, intype[i]);	/* Input type (SVIDEO/CVIDEO).*/
			CORE_attr_set(h_hbd, ATTR_stream_type, streamtype[i]);	/* Mpeg multiplex type.*/
			CORE_attr_set(h_hbd, ATTR_video_chan, inchan[i]);	/* Input channel.*/
			CORE_attr_set(h_hbd, ATTR_mpeg, mpeg[i]);	/* Enumerated picture size and mpeg type.*/
			CORE_attr_set(h_hbd, ATTR_video_bitrate, vbitrate[i]);	/* Video bit rate in bps.*/
			CORE_attr_set(h_hbd, ATTR_video_vbr_enable, vbron[i]);	/* Variable bit rate enable.*/
			CORE_attr_set(h_hbd, ATTR_av_select, avsel[i]);	/* Mux A/V select.*/
			CORE_attr_set(h_hbd, ATTR_video_m, gop[i]);	/* GOP size.*/
			CORE_attr_set(h_hbd, ATTR_video_n, ipdist[i]);	/* IP distance.*/
			CORE_attr_set(h_hbd, ATTR_audio_inampgain, ampgain[i]);	/* Manual audio gain*/
			CORE_attr_set(h_hbd, ATTR_audio_agc_onoff, agcon[i]);	/* AGC on off*/
			CORE_attr_set(h_hbd, ATTR_audio_agc_sens, agcsens[i]);	/* AGC sensitivity*/
			CORE_gpio_write(h_hbd, gpio[i]);	/* GPIO*/
			/* Program the settings that are fixed.*/
			CORE_attr_set(h_hbd, ATTR_video_brightness, DFLT_BRIGHTNESS);	/* Luminance brightness.*/
			CORE_attr_set(h_hbd, ATTR_video_contrast, DFLT_CONTRAST);	/* Luminance contrast.*/
			CORE_attr_set(h_hbd, ATTR_video_saturation, DFLT_SATURATION);	/* Chrominance saturation.*/
			CORE_attr_set(h_hbd, ATTR_audio_bitrate, DFLT_AUDIO_BITRATE);	/* Audio bit rate.*/
			CORE_attr_set(h_hbd, ATTR_audio_pcm_resolution,	DFLT_AUDIO_PCM_RES); /* Enumerated PCM resolution.*/

			/* Activate the board's default settings.*/
			CORE_board_reset(h_hbd);

			/* Initialize the board's wait queue, which is a list of sleeping processes waiting to read/write mpeg data.*/
			init_waitqueue_head(&stream_q[i]);
			/* Initialize the board's wait queue, which is a list of sleeping processes waiting to video field/frame sync.*/
			init_waitqueue_head(&video_q[i]);

			/* Install interrupt handler, if irq is non-zero.*/
			if (irq[i] > 0)	{
				/* Fast interrupt*/
#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 18)
				if (request_irq(irq[i], interrupt_handler, SA_INTERRUPT | SA_SHIRQ, devname[i], &board_number[i])) {
#else
#if LINUX_VERSION_CODE < KERNEL_VERSION(4, 0, 0)
				if (request_irq(irq[i], interrupt_handler, IRQF_SHARED | IRQF_DISABLED, devname[i], &board_number[i])) {
#else
				if (request_irq(irq[i], interrupt_handler, IRQF_SHARED, devname[i], &board_number[i])) {
#endif


#endif
					/* Report the successful installation.*/
					printk(KERN_INFO "x16(%d) interrupt %d assignment failed"
					       " - polling used.\n", i, irq[i]);
					irq[i] = 0;
				} else {
					printk(KERN_INFO "x16(%d) interrupt %d for device %s"
					       "installed.\n", i, irq[i], devname[i]);
				}
			}
			/* Indicate board has been successfully installed.*/
			board_installed[i] = TRUE;
#if defined(SX16_V4L_BUILD)
			{
				struct sx16_v4l *v4l;
				v4l = G_v4l[i];
				v4l->video_q.hbd = hbd[i];
				v4l->hbd = hbd[i];
			}
#endif
			/* Report the successful installation.*/
			printk(KERN_INFO "x16(%d) installed. %p\n", i, hbd[i]);

		}
	}

	/* Indicate success/fail.*/
	return rtnval;
}


/* Reset all installed boards and terminate the driver.*/

static void remove_driver(void)
{
	int i;
	/* Reset and release each installed board. For each possible board ...*/
	for (i = 0; i < num_boards; i++) {
		/* If the board is installed...*/
		if (board_installed[i])	{
			if (irq[i] > 0) {
				free_irq(irq[i], &board_number[i]);
			}

			/* Shutdown the board.*/
			CORE_board_shutdown(hbd[i]);
			hbd[i] = NULL;

			/* Release the board's resources.*/
			release_resources(i);

#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 5, 0)
			if (start_thread_flag[i]) {
				stop_kthread(&start_thread_tcb[i]);
				start_thread_flag[i] = 0;
			}
#else
			if (start_thread_tcb[i]) {
				kthread_stop(start_thread_tcb[i]);
				start_thread_tcb[i] = NULL;
			}
#endif
			if (dmabuf[i]) {
				kfree(dmabuf[i]);
				dmabuf[i] = NULL;
			}
#if defined(SX16_V4L_BUILD)
			if (model_num == 616) {
				unregister_video(i);
			}
#endif
		}
	}
	/* Terminate the driver.*/
	unregister_chrdev(major, "sx16");
}

/* Construct a list of pointers to the pci_dev structures for all model 616 boards.*/
/* Populates s616_list[] and returns the count of model 616 boards.*/

#define SAA7146_VENDORID	0x1131	/* Philips.*/
#define SAA7146_DEVICEID	0x7146	/* SAA7146 PCI bridge.*/
#define S616_SUBVENDORID	0x6000	/* Sensoray.*/
#define S616_SUBDEVICEID	0x0616	/* Model 616.*/

static unsigned int find_616s(struct pci_dev **devlist)
{
	int				nboards		= 0;
	struct pci_dev	*dev		= NULL;
	unsigned short	sub_vendor;
	unsigned short	sub_device;
	printk(KERN_DEBUG "sx16driver: nosubid %d\n", nosubid);
	/* Repeat until we have found the maximum number of
	   supported model 616 boards ...*/
#ifdef TEST_PCI_616	
	nosubid = 1;
#endif
	do {
		/* Find next board that has the correct main ids.*/
#ifdef TEST_PCI_616
		dev = pci_get_device(0x8086, 0x2f02, dev);//SAA7146_VENDORID, SAA7146_DEVICEID, dev);
#else
		dev = pci_get_device(SAA7146_VENDORID, SAA7146_DEVICEID, dev);
#endif
		
		/* Exit search loop if no more potential 616 boards are present.*/
		if (!dev)
			break;
		/* Fetch the PCI sub-ids from the candidate board.*/
		pci_read_config_word(dev, PCI_SUBSYSTEM_VENDOR_ID, &sub_vendor);
		pci_read_config_word(dev, PCI_SUBSYSTEM_ID, &sub_device);

		/* If sub-ids match, add board to the list.
		   If sub ids are not valid add board to list.*/
		if (nosubid || ((sub_vendor == S616_SUBVENDORID) && (sub_device == S616_SUBDEVICEID))) {
			*devlist++ = dev;
			nboards++;
		}
	}
	while (nboards < MAX_BOARDS);

	/* Return the number of 616 boards that were found.*/
	return nboards;
}


/* Install the driver and all declared boards.	If no "io=" parameters are specified then it is assumed*/
/* we are installing model 616 boards, otherwise the "io=" parameters are assumed to explicitly declare*/
/* the base I/O addresses of all model 516 boards.*/

static int install_driver(void)
{
	int i;

	/* Count the number of 516 boards suggested by the insmod "io=" parameters.*/
	for (num_boards = 0; (io[num_boards] != 0) && (num_boards < MAX_BOARDS); num_boards++);

	/* If no model 516 boards were declared ...*/
	if (num_boards == 0) {
		/* Assume we are installing 616s:
		   scan the PCI bus and construct a list of all model 616 boards.*/
		model_num = 616;
		printk(KERN_INFO "Searching for 616s.\n");
		num_boards = find_616s(s616_dev);
	}

	/* Terminate driver if no boards are available.*/
	if (num_boards == 0) {
		printk(KERN_ERR "sx16: Missing 'io=' module parameter for 516s,"
		       " and no 616s(PCI bus) were found.\n");
		return INSTALL_FAIL;
	}

	/* Attempt to register the device with the kernel,
	   and if unsuccessful, report the problem and terminate the driver.*/
	/* sx16 operations done with /dev/sx16aX.
	   video4linux done through /dev/videoX*/
	if (register_chrdev(major, "sx16", &sx16_fops)) {
		printk(KERN_ERR "Unable to get major number %i for X16.\n", major);
		return INSTALL_FAIL;
	}

	printk(KERN_INFO "Initializing boards.\n");

#if defined(SX16_V4L_BUILD)
	/* Registering video devices.  For each X16 board ...*/
	if (model_num == 616) {
		for (i = 0; i < num_boards; i++) {
			if (register_video(i)) {
				remove_driver();
				printk(KERN_ERR
				       "Unable to register video for board (major number: %i).\n", major);
				return INSTALL_FAIL;
			}
		}
	}
#endif
	/* Install all declared boards, terminate driver if unsuccessful.
	   For each X16 board ...*/
	for (i = 0; i < num_boards; i++) {
		/* Attempt to install the board, and if not successful ...*/
		if (init_board(i) == INSTALL_FAIL) {
			/* Uninstall all installed boards
			   and terminate the driver.*/
			remove_driver();
			printk(KERN_ERR
			       "sx16: initialize board failed (major number: %i).\n",
			       major);
			return INSTALL_FAIL;
		}
	}

	/* Make public the function that supplies "proc" info to the system.*/
#if LINUX_VERSION_CODE < KERNEL_VERSION(3, 10, 0)
	create_proc_read_entry("sx16", 0, 0, sx16_read_procmem, NULL);
#else
	proc_create_data("sx16", 0, 0, &sx16_procmem_fops, NULL);
#endif
	return INSTALL_PASS;
}


/* INSTALL METHOD.  This is called when the module is loaded.*/
/* Register the module and install all declared boards.*/
static int __init init_module_sx16(void)
{
	int rtnval;
	int j;
	printk(KERN_INFO "\nInstalling x16 driver...(%s)\n", GetRelease());
	/* initialize our global parameters */

	sema_init(&sx16_devices_lock, 1);
	down(&sx16_devices_lock);
	for (j = 0; j < MAX_BOARDS; j++) {
		hbd[j] = NULL;
#if defined(SX16_V4L_BUILD)
		G_v4l[j] = NULL;
#endif
		open_count_mpeg[j] = 0;
		open_count_v4l[j] = 0;
		board_number[j] = j;
#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 5, 0)
		poll_thread_exists[j] = FALSE;
#else
		poll_thread_tcb[j] = NULL;
#endif
		start_thread_flag[j] = 0;
		board_installed[j] = FALSE;
		sprintf(poll_thread_name[j], "sx16_pthread_%d", j);
		sprintf(start_thread_name[j], "sx16_sthread_%d", j);
		sprintf(devname[j], "%s%d", DEVNAME_ROOT, j);
		dmabuf[j] = NULL;
		PacketCountRx[j] = 0;
	}



	printk(((rtnval = install_driver()) == INSTALL_PASS) ?
	       "sx16 driver successfully installed!\n\n" :
	       "sx16 driver installation failed!\n\n");
	up(&sx16_devices_lock);
	return rtnval;
}


/* REMOVE METHOD.  This is called when the module is unloaded.*/
/* Reset and release each installed board, terminate driver.*/

static void __exit cleanup_module_sx16(void)
{
	remove_driver();
	remove_proc_entry("sx16", NULL);
	printk(KERN_INFO "sx16 driver removed!\n");
}


/* Does a kernel print of the supplied text string.*/
/**/
void SX16MOD_text_message(char *txt)
{
	printk(KERN_INFO "sx16: ");
	printk(KERN_INFO "%s", txt);
}

/* Does a kernel print of an integer as hex number.*/
/**/
void SX16MOD_hex_message(int digits, int num)
{
	printk(KERN_INFO "SX16MOD_hex_message ");
	switch (digits) {
	case 1:	 printk(KERN_INFO "%1.1X", num); break;
	case 2:	 printk(KERN_INFO "%2.2X", num); break;
	case 3:	 printk(KERN_INFO "%3.3X", num); break;
	case 4:	 printk(KERN_INFO "%4.4X", num); break;
	case 5:	 printk(KERN_INFO "%5.5X", num); break;
	case 6:	 printk(KERN_INFO "%6.6X", num); break;
	case 7:	 printk(KERN_INFO "%7.7X", num); break;
	case 8:	 printk(KERN_INFO "%8.8X", num); break;
	default: printk(KERN_INFO "%X",	  num); break;
	}
}

/* sleep stream queue for board*/
/* timeout = milliseconds(careful with HZ resolution,*/
/*			   ie, use a large number for more accuracy)*/
void SX16MOD_sleepq(void *board, int timeout)
{
	int board_num = handle_to_boardnum((HBD) board);
#if LINUX_VERSION_CODE > KERNEL_VERSION(2, 5, 0)
	DEFINE_WAIT(wait);
#endif
	if (board_num < 0)
		return;

	if (timeout == 0) {
#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 5, 0)
		interruptible_sleep_on(&stream_q[board_num]);
#else
		/* function above deprecated on 2.6 kernels.*/
		prepare_to_wait(&stream_q[board_num], &wait, TASK_INTERRUPTIBLE);
		schedule();
		finish_wait(&stream_q[board_num], &wait);
#endif
	} else {
#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 5, 0)
		interruptible_sleep_on_timeout(&stream_q[board_num], timeout);
#else
		prepare_to_wait(&stream_q[board_num], &wait, TASK_INTERRUPTIBLE);
		schedule_timeout(timeout);
		finish_wait(&stream_q[board_num], &wait);
#endif
	}
	return;
}

/* Sleep video sync queue for board.*/
static void video_sleepq(void *board, int timeout)
{
	int board_num = handle_to_boardnum((HBD) board);
#if LINUX_VERSION_CODE > KERNEL_VERSION(2, 5, 0)
	DEFINE_WAIT(wait);
#endif
	if (board_num < 0 || timeout <= 0)
		return;

	/* Linux doesn't accommodate less
	   than MIN_SLEEP_TIME millisecond sleep time.*/
	if (timeout < MIN_SLEEP_TIME)
		timeout = MIN_SLEEP_TIME;
	/* Convert msec to timer's units.*/
	timeout = timeout * HZ / 1000;
	/*printk(KERN_INFO "timeout = %d HZ = %d\n", timeout, HZ);*/
#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 5, 0)
	interruptible_sleep_on_timeout(&video_q[board_num], timeout);
#else
	prepare_to_wait(&video_q[board_num], &wait, TASK_INTERRUPTIBLE);
	schedule_timeout(timeout);
	finish_wait(&video_q[board_num], &wait);
#endif
}


/* Suspend the calling process for specified time interval.*/
void SX16MOD_sleep(int milliseconds)
{
#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 5, 0)
	wait_queue_head_t sleep_q;		/* sleep wait queue.*/
	/* If delay time is zero, simply schedule;
	   no delay is required (or desired) by caller.*/
	if (milliseconds == 0) {
		schedule();
		return;
	}
	/* Linux doesn't accommodate less than MIN_SLEEP_TIME
	   millisecond sleep time.*/

	if (milliseconds < MIN_SLEEP_TIME)
		milliseconds = MIN_SLEEP_TIME;

	/* Fall asleep for the specified milliseconds.*/
	init_waitqueue_head(&sleep_q);

	/* Fall asleep for the specified milliseconds.*/
	interruptible_sleep_on_timeout(&sleep_q, (milliseconds * HZ) / 1000);
#else
	if (milliseconds == 0) {
		schedule();
		return;
	}
	msleep_interruptible(milliseconds);
#endif
}


/* Return number of milliseconds elapsed since midnight.*/

unsigned __int64 SX16MOD_time_now()
{
	struct timeval tv;
	/* Get the current time of day.*/
	do_gettimeofday(&tv);

	/* Compute and return the elapsed milliseconds since midnight.*/
	return (unsigned long long)tv.tv_sec * (unsigned long long)1000 + (unsigned long long)(tv.tv_usec / 1000);
}



/* Read/write 16-bit I/O port.*/

unsigned short SX16MOD_read_io16(int addr)
{
#ifdef TEST_PCI_616
	return 0;
#else
	return inw(addr);
#endif
}

void SX16MOD_write_io16(int addr, unsigned short data)
{
#ifdef TEST_PCI_616
	return;
#else
	outw(data, addr);
#endif
}



/* Read/write 32-bit memory port.*/

unsigned long SX16MOD_read_mem32(unsigned long addr)
{
#ifdef TEST_PCI_616
	return 0;
#endif

#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 8)
	return readl(addr);
#else
	return readl((volatile void __iomem *) addr);
#endif
}


void SX16MOD_write_mem32(unsigned long addr, unsigned long data)
{
#ifdef TEST_PCI_616
	return;
#endif
	
#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 8)
	writel(data, addr);
#else
	writel(data, (volatile void __iomem *) addr);
#endif
}

/* Input data block of word values from hardware register.*/

void SX16MOD_block_read_io16(int addr, unsigned short *buf, int nwords)
{
#ifdef TEST_PCI_616
	return;
#endif	
	insw(addr, buf, nwords);
}


/* Output data block of word values to hardware register.*/

void SX16MOD_block_write_io16(int addr, unsigned short *buf, int nwords)
{
#ifdef TEST_PCI_616
	return;
#endif	
	outsw(addr, buf, nwords);
}


/* A 'malloc' implementation.*/

void *SX16MOD_malloc(int size)
{
	return vmalloc(size);
}


/* A 'malloc' implementation.*/

void SX16MOD_alloc_dma(HBD board, int size, unsigned long **cpu_addr, unsigned long *dma_handle)
{
	int board_num = handle_to_boardnum(board);
	*cpu_addr = pci_alloc_consistent(s616_dev[board_num], size, (dma_addr_t *)  dma_handle);
	return;
}

void SX16MOD_free_dma(HBD board, int size, unsigned long *cpu_addr, unsigned long dma_handle)
{
	int board_num = handle_to_boardnum(board);
	pci_free_consistent(s616_dev[board_num], size, cpu_addr, (dma_addr_t) dma_handle);
}


/* A 'free' implementation.*/

void SX16MOD_free(void *obj)
{
	if (obj)
		vfree(obj);
}


/* A 'memmove' implementation.*/

void SX16MOD_memmove(char *dst, char *src, int size)
{
	memmove(dst, src, size);
}


/* A 'memcmp' implementation.*/

int SX16MOD_memcmp(void *s1, void *s2, int size)
{
	return memcmp(s1, s2, size);
}

/* Mutex functions.*/

void *SX16MOD_create_mutex(void)
{
	struct semaphore *psem;

	psem = (struct semaphore *)kmalloc(sizeof(struct semaphore),	 GFP_KERNEL | __GFP_DMA);

	if (!psem)
		return NULL;

	sema_init(psem, 1);

	return psem;
}

void SX16MOD_delete_mutex(void *mutex_object)
{
	if (mutex_object)
		kfree(mutex_object);
}

int SX16MOD_acquire_mutex(void *mutex_object)
{
	int res;
	res = down_interruptible((struct semaphore *)mutex_object);
	if (res)
		return -1;
	else
		return 0;
}

void SX16MOD_acquire_mutex_nonint(void *mutex_object)
{
    int rc = 0;
	if (mutex_object)
		rc = down_interruptible((struct semaphore *)mutex_object);
    if (rc != 0)
        printk(KERN_INFO "failed mutex\n");
}

void SX16MOD_release_mutex(void *mutex_object)
{
	if (mutex_object)
		up((struct semaphore *)mutex_object);
}

/* private function called from CORE on board_close*/
/* releases the threads.  called from CORE on driver close/release*/
void SX16MOD_release_threads(HBD hboard)
{
	int board = handle_to_boardnum((HBD) hboard);
	/* Kill the poll thread.*/
#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 5, 0)
	if (poll_thread_exists[board]) {
		stop_kthread(&poll_thread_tcb[board]);
		poll_thread_exists[board] = FALSE;
	}
	if (start_thread_flag[board]) {
		stop_kthread(&start_thread_tcb[board]);
		start_thread_flag[board] = 0;
	}
#else
	if (poll_thread_tcb[board]) {
		kthread_stop(poll_thread_tcb[board]);
		poll_thread_tcb[board] = NULL;
	}
	if (start_thread_tcb[board]) {
		kthread_stop(start_thread_tcb[board]);
		start_thread_tcb[board] = NULL;
	}
#endif
	return;
}


/* Get "file handle" to decompressor, or 0 if open fails.*/
void *SX16MOD_open_microcode(int microcode_id, int streamtype,
			     int audio_fmt)
{

	switch (microcode_id) {
	case MPEG_RECORD:
		if (streamtype == SX16_STREAMTYPE_TS) {
#ifdef SX16_ENABLE_TS
			return open_decompressor(record_microcode_ts, SX16MOD_malloc);
#else
#warning "NO TS support.  Not a warning if smaller binary desired."
			printk(KERN_INFO "SX16: TS not compiled into driver!\n");
			return open_decompressor(record_microcode, SX16MOD_malloc);
#endif
		} else {
#ifdef AUDIO_EXTRA_FMTS_MP3
#warning "Compiling with MP3 support"
#warning "Please check licensing for MP3"
			if (audio_fmt == 2) {
				printk(KERN_INFO "sx16: using 083 firmware for MP3 audio\n");
				return open_decompressor(record_microcode_083, SX16MOD_malloc);
			}
#endif
#ifdef AUDIO_EXTRA_FMTS_AC3
#warning "Compiling with AC3 support"
#warning "Please check licensing with Dolby Labs if using AC3"
			if (audio_fmt == 3) {
				printk(KERN_INFO "sx16: using 600 firmware for AC3 audio\n");
				return open_decompressor(record_microcode_600, SX16MOD_malloc);
			}
#endif
			return open_decompressor(record_microcode, SX16MOD_malloc);
		}
	case MPEG_PLAY:
		if (streamtype == SX16_STREAMTYPE_VES) {
			return open_decompressor(play_microcode_ves, SX16MOD_malloc);
		} else {
			return open_decompressor(play_microcode, SX16MOD_malloc);
		}

	default:
		return 0;
	}
}

/* Read specified number of bytes of decompressed data into our buffer.*/
/* Return number of bytes actually returned; "end of data" condition is*/
/* indicated if we get less data than we asked for.*/

int SX16MOD_read_microcode(void *handle, void *databuf, int desired_byte_count)
{
	return read_decompressor(handle, databuf, desired_byte_count);
}

/* Close decompressor.*/

void SX16MOD_close_microcode(void *handle)
{
	close_decompressor(handle, vfree);
}


#ifdef SX16_BUILD_SRAY
#define REPORT_SIZE 1024

int RTCP_SSCP_Report(HBD board, void *val, int cmd)
{
	int retval;
	int board_num;
	int rtphandle = CORE_get_rtp_handle(board);
	board_num = handle_to_boardnum(board);
	if (board_num == -1)
		return -EFAULT;

	switch (cmd) {
	case SX16_IOC_GET_RTCP_SR:
		if (!ACCESS_OK(VERIFY_WRITE, val, REPORT_SIZE))
			retval = -EFAULT;
		else
			retval = GetRTCP_SR(rtphandle, (unsigned char *) val);
		break;

	case SX16_IOC_GET_RTCPA_SR:
		if (!ACCESS_OK(VERIFY_WRITE, val, REPORT_SIZE))
			retval = -EFAULT;
		else
			retval = GetRTCPA_SR(rtphandle, (unsigned char *) val);
		break;

	case SX16_IOC_GET_RTCP_RR:
		if (!ACCESS_OK(VERIFY_WRITE, val, REPORT_SIZE))
			retval = -EFAULT;
		else
			retval = GetRTCP_RR(rtphandle, (unsigned char *) val);
		break;

	case SX16_IOC_PUT_RTCP_RR:
		if (!ACCESS_OK(VERIFY_READ, val, REPORT_SIZE))
			retval = -EFAULT;
		else
			retval = PutRTCP_RR(rtphandle, (unsigned char *) val);
		break;

	case SX16_IOC_PUT_RTCP_SR:
		if (!ACCESS_OK(VERIFY_READ, val, REPORT_SIZE))
			retval = -EFAULT;
		else
			retval = PutRTCP_SR(rtphandle, (unsigned char *) val);
		break;
	case SX16_IOC_GET_SSCP_SR:
		if (!ACCESS_OK(VERIFY_WRITE, val, REPORT_SIZE))
			retval = -EFAULT;
		else {
			write_param_t writeParam;
			CORE_get_write_params(hbd[board_num], &writeParam);
			retval = GetSSCP_SR(board_num, (unsigned char *) val, writeParam.PacketCount, writeParam.PacketCount * 1442, do_gettimeofday);
		}
		break;

	case SX16_IOC_GET_SSCP_RR:
		if (!ACCESS_OK(VERIFY_WRITE, val, REPORT_SIZE))
			retval = -EFAULT;
		else {
			write_param_t writeParam;
			CORE_get_write_params(hbd[board_num], &writeParam);
			retval = GetSSCP_RR(board_num, (unsigned char *) val, writeParam.SequenceNumber, writeParam.LostPackets);
		}
		break;
	case SX16_IOC_PUT_SSCP_RR:
		if (!ACCESS_OK(VERIFY_READ, val, REPORT_SIZE))
			retval = -EFAULT;
		else
			retval = PutSSCP_RR(board_num, (unsigned char *) val);
		break;

	case SX16_IOC_PUT_SSCP_SR:
		if (!ACCESS_OK(VERIFY_READ, val, REPORT_SIZE))
			retval = -EFAULT;
		else
			retval = PutSSCP_SR(board_num, (unsigned char *) val);
		break;

	default:
		retval = -EINVAL;
		break;

	}
	return retval;
}
#endif

/*
 * The following function is run in a separate thread.
 * It starts the decoder(uploads ucode, etc)
 * while the main thread may either block (when playing back a file),
 *	or return immediately discarding the data while the decoder is not ready
 *(used when connecting to network sream).
 * start_thread_flag is used for communication between main driver thread
 *    and "start" thread
 * 0: initial (idle) value
 * 1: main thread started "start" thread (set by main thread)
 * 2: "start" thread ailed (set by "start" thread)
 * 3: "start" thread is done (set by "start" thread)
 */
#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 5, 0)
void start_dec_thread(kthread_t *kthread)
#else
int start_dec_thread(void *vp)
#endif
{
	/* Setup the thread environment.*/
#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 5, 0)
	long board = kthread->board;
	HBD	h_hbd = hbd[board];
	init_kthread(kthread, kthread->arg);
#else
	long board = (long) vp;
	HBD h_hbd = hbd[board];
	wait_queue_head_t start_sleep_q; /* sleep wait queue.*/
	DEFINE_WAIT(wait);
	init_waitqueue_head(&start_sleep_q);
#endif

	CORE_start_decode(h_hbd);		/* upload ucode and start*/

	if (!CORE_stream_is_decoding(h_hbd)) {
		start_thread_flag[board] = 2;		/* failed*/
	} else {
#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 5, 0)
		interruptible_sleep_on_timeout(&kthread->queue, 10);
#else
		prepare_to_wait(&start_sleep_q, &wait, TASK_INTERRUPTIBLE);
		schedule_timeout(10);
		finish_wait(&start_sleep_q, &wait);
#endif
		start_thread_flag[board] = 3;		/* succeded*/

	}

#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 5, 0)
    while (!kthread->terminate) {
	    interruptible_sleep_on_timeout(&kthread->queue, 5);
    }
	exit_kthread(kthread);
#else
    while (!kthread_should_stop()) {
       prepare_to_wait(&start_sleep_q, &wait, TASK_INTERRUPTIBLE);
       schedule_timeout(5);
       finish_wait(&start_sleep_q, &wait);
    }
    return 0;
#endif
}

/*if skip_data != 0 this function blocks until decoder
  is ready to accept data*/
/* otherwise returns immediately.
   Return value is non-zero while decoder is not ready.*/

static int start_decoder(int board, int skip_data)
{
	HBD	h_board = hbd[board];
	while (TRUE) {
		/* Start the encoder if it is not already running.*/
		if (!CORE_stream_is_decoding(h_board)) {
			/* not decoding anymore so reset the sensoray write
			   state machine in the core*/
			CORE_reset_write_step(h_board);
			/*if board is not ready to decode*/
			if ((start_thread_flag[board] == 0) ||
			    (start_thread_flag[board] == 2)) {
				/*start thread was not started or failed*/
#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 5, 0)
				start_thread_tcb[board].arg = start_thread_name[board];
				start_thread_tcb[board].board = board;

				start_kthread(start_dec_thread, &start_thread_tcb[board]); /*start the "start thread"*/
#else
				start_thread_tcb[board] = kthread_run(start_dec_thread, (void *)(long)board, start_thread_name[board]);
				if (start_thread_tcb[board] < 0) {
					start_thread_tcb[board] = NULL;
					printk(KERN_ERR "Unable to start kernel thread for X16.\n");
					return 0;
				}
#endif
				start_thread_flag[board] = 1;  /*set flag to "started the start thread"*/
			} else if (start_thread_flag[board] != 1) {
				/*SM was somehow stopped*/
#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 5, 0)
				stop_kthread(&start_thread_tcb[board]);
#else
				if (start_thread_tcb[board]) {
					kthread_stop(start_thread_tcb[board]);
					start_thread_tcb[board] = NULL;
				}
#endif
				start_thread_flag[board] = 0;	/*restart it*/
			}
			if (skip_data)
				return 1;
			else {
				SX16MOD_sleep(10);
				continue;
			}
		} else if (start_thread_flag[board]) {
			if (start_thread_flag[board] != 3) {
				/*if "start" thread is not done yet*/
				if (skip_data)
					return 1;
				else {

					SX16MOD_sleep(10);
					continue;
				}
			} else {
				/*"start" thread signaled that it's done*/
#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 5, 0)
				stop_kthread(&start_thread_tcb[board]);
#else
				if (start_thread_tcb[board]) {
					kthread_stop(start_thread_tcb[board]);
					start_thread_tcb[board] = NULL;
				}
#endif
				/*thread terminated*/
				start_thread_flag[board] = 0;
			}
		}
		break;
	}
	return 0;
}


unsigned long SX16MOD_get_jiffies(void)
{
	return jiffies;
}

/* do_gettimeofday function*/
void SX16MOD_do_gettimeofday(void *tval)
{
	struct timeval *tv = (struct timeval *) tval;
	do_gettimeofday(tv);
}

/* do_gettimeofday function*/
void SX16MOD_do_gettimeofday_secs(long *tval)
{
	struct timeval tmp ;
	do_gettimeofday(&tmp);
	*tval = tmp.tv_sec;
	;
}
void SX16MOD_do_gettimeofday_usecs(long *tval)
{
	struct timeval tmp ;
	do_gettimeofday(&tmp);
	*tval = tmp.tv_usec;
	;
}


/* clock hz function*/
int SX16MOD_get_hz(void)
{
	return HZ;
}


void SX16MOD_lpt(unsigned char data)
{
	outw(data, 0x378);
}



/* for uncompressed capture mode.   gets current physical address of virtual address using page lookup*/

unsigned long SX16MOD_getphys(void *virt_addr)
{
#if LINUX_VERSION_CODE > KERNEL_VERSION(2, 6, 3)
	pgd_t *pgd;
	pmd_t *pmd;
	pte_t *pte;
	unsigned long pa;
#if LINUX_VERSION_CODE > KERNEL_VERSION(2, 6, 10)
	pud_t *pud;
#endif
#if LINUX_VERSION_CODE >= KERNEL_VERSION(4, 12, 0)
	p4d_t *p4d;
#endif
	pgd = pgd_offset(current->mm, (unsigned long)virt_addr);
	if (pgd_none(*pgd) || pgd_bad(*pgd)) {
		printk(KERN_ERR "getphys: pgd bad or none: va %lx\n", (unsigned long)virt_addr);
		return 0;
	}
#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 11)
	pmd = pmd_offset(pgd, (unsigned long)virt_addr);
#elif LINUX_VERSION_CODE < KERNEL_VERSION(4, 12, 0)
	pud = pud_offset(pgd, (unsigned long) virt_addr);
	pmd = pmd_offset(pud, (unsigned long) virt_addr);
#else
	p4d = p4d_offset(pgd, (unsigned long) virt_addr);
	pud = pud_offset(p4d, (unsigned long) virt_addr);
	pmd = pmd_offset(pud, (unsigned long) virt_addr);
#endif
	
	if (pmd_none(*pmd) || pmd_bad(*pmd)) {
		printk(KERN_ERR "getphys: pmd bad or none: va %lx\n", (unsigned long)virt_addr);
		return 0;
	}
#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 5, 0)
	pte = pte_offset(pmd, (unsigned long)virt_addr);
#else
	pte = pte_offset_kernel(pmd, (unsigned long)virt_addr);
#endif
	if (!pte_present(*pte)) {
		printk(KERN_ERR "getphys: wrong virtual address %lx\n", (unsigned long)virt_addr);
		return(0);
	}
	pa = (*(int *)pte) & ~0xfff;
	return((pa) + ((unsigned long)virt_addr & 0x0FFF));
#else
	return(0);
#endif /*LINUX_VERSION_CODE > KERNEL_VERSION(2, 6, 3)*/
}


/* get kernel specific PAGE_SIZE*/
unsigned long SX16MOD_get_PAGESIZE()
{
	return PAGE_SIZE;   /* for 616 must be 4096 !!!*/
}

#if defined(SX16_V4L_BUILD)
#if LINUX_VERSION_CODE >= KERNEL_VERSION_NOINODE
static const struct v4l2_file_operations v16_fops = {
#else
static 
#if LINUX_VERSION_CODE > KERNEL_VERSION(2, 6, 16)
const 
#endif
struct file_operations v16_fops = {
#endif
	.owner	  = THIS_MODULE,
	.open	  = video_open,
	.release  = video_release,
#if LINUX_VERSION_CODE >= KERNEL_VERSION_NOINODE	
	.unlocked_ioctl = video_ioctl,
#else
	.ioctl = video_ioctl,
#endif
	//.read	  = video_read,
	//.write  = video_write,
	.mmap	  = video_mmap,
	.poll	  = video_poll,
#if LINUX_VERSION_CODE < KERNEL_VERSION_NOINODE	
	.llseek	  = no_llseek,
#endif
};

#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 20)
/* kernels 2.6.20 and above have a much better way of doing this */
static struct v4l2_tvnorm tvnorms[] = {
	{
		.name      = "NTSC",
		.id        = V4L2_STD_NTSC,
	}, {
		.name      = "PAL",
		.id        = V4L2_STD_PAL,
	}, {
		.name      = "NTSC-M",
		.id        = V4L2_STD_NTSC_M,
	}, {
		.name      = "NTSC-M-JP",
		.id        = V4L2_STD_NTSC_M_JP,
	}, {
		.name      = "NTSC-M-KR",
		.id        = V4L2_STD_NTSC_M_KR,
	}, {
		.name      = "PAL-BG",
		.id        = V4L2_STD_PAL_BG,
	}, {
		.name      = "PAL-G",
		.id        = V4L2_STD_PAL_G,
	}, {
		.name      = "PAL-H",
		.id        = V4L2_STD_PAL_H,
	}, {
		.name      = "PAL-I",
		.id        = V4L2_STD_PAL_I,
	}, {
		.name      = "PAL-DK",
		.id        = V4L2_STD_PAL_DK,
	}

};

#endif

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 27)
static const struct v4l2_ioctl_ops v16_ioctl_ops = {
	.vidioc_querycap = vidioc_querycap,
	.vidioc_enum_fmt_vid_cap = vidioc_enum_fmt_cap,
	.vidioc_g_fmt_vid_cap = vidioc_g_fmt_cap,
	.vidioc_try_fmt_vid_cap = vidioc_try_fmt_cap,
	.vidioc_s_fmt_vid_cap = vidioc_s_fmt_cap,
	.vidioc_reqbufs = vidioc_reqbufs,
	.vidioc_querybuf = vidioc_querybuf,
	.vidioc_qbuf = vidioc_qbuf,
	.vidioc_dqbuf = vidioc_dqbuf,
	.vidioc_s_std = vidioc_s_std,
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 32)
	.vidioc_g_std         = vidioc_g_std,
#endif
	.vidioc_enum_input = vidioc_enum_input,
	.vidioc_g_input = vidioc_g_input,
	.vidioc_s_input = vidioc_s_input,
	.vidioc_queryctrl = vidioc_queryctrl,
	.vidioc_g_ctrl = vidioc_g_ctrl,
	.vidioc_s_ctrl = vidioc_s_ctrl,
	.vidioc_streamon = vidioc_streamon,
	.vidioc_streamoff = vidioc_streamoff,
#ifdef CONFIG_VIDEO_V4L1_COMPAT
	.vidiocgmbuf = vidioc_cgmbuf,
#endif
};
#endif
	
	
static struct video_device v16_video_template = {
	.name	  = "UNSET",
	.name = "sx16v",

	.fops	  = &v16_fops,
	.minor	  = -1,
	.release = video_device_release,
#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 27)
	.type = VID_TYPE_CAPTURE,
	.vidioc_querycap = vidioc_querycap,
#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 27)
	.vidioc_enum_fmt_cap = vidioc_enum_fmt_cap,
	.vidioc_g_fmt_cap = vidioc_g_fmt_cap,
	.vidioc_try_fmt_cap = vidioc_try_fmt_cap,
	.vidioc_s_fmt_cap = vidioc_s_fmt_cap,
#else
	.vidioc_enum_fmt_vid_cap = vidioc_enum_fmt_cap,
	.vidioc_g_fmt_vid_cap = vidioc_g_fmt_cap,
	.vidioc_try_fmt_vid_cap = vidioc_try_fmt_cap,
	.vidioc_s_fmt_vid_cap = vidioc_s_fmt_cap,
#endif
	.vidioc_reqbufs = vidioc_reqbufs,
	.vidioc_querybuf = vidioc_querybuf,
	.vidioc_qbuf = vidioc_qbuf,
	.vidioc_dqbuf = vidioc_dqbuf,
#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 20)
	.vidioc_s_std = vidioc_s_std_old,
#else
	.vidioc_s_std = vidioc_s_std,
#endif
	.vidioc_enum_input = vidioc_enum_input,
	.vidioc_g_input = vidioc_g_input,
	.vidioc_s_input = vidioc_s_input,
	.vidioc_queryctrl = vidioc_queryctrl,
	.vidioc_g_ctrl = vidioc_g_ctrl,
	.vidioc_s_ctrl = vidioc_s_ctrl,
	.vidioc_streamon = vidioc_streamon,
	.vidioc_streamoff = vidioc_streamoff,
#ifdef CONFIG_VIDEO_V4L1_COMPAT
	.vidiocgmbuf = vidioc_cgmbuf,
#endif
#define HAVE_CURRENT_NORM
	.current_norm = V4L2_STD_NTSC_M,
#else
	.ioctl_ops = &v16_ioctl_ops,	
#endif

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 20)
	.tvnorms = V4L2_STD_NTSC | V4L2_STD_PAL,
#else
	.tvnorms              = tvnorms,
	.tvnormsize           = ARRAY_SIZE(tvnorms),
#endif

};


#ifndef HAVE_V4L2			      /* v4l2 does not supported: add some things...*/

struct video_device *video_device_alloc(void)
{
	struct video_device *vfd;

	vfd = kmalloc(sizeof(*vfd), GFP_KERNEL);
	if (NULL == vfd)
		return NULL;
	memset(vfd, 0, sizeof(*vfd));
	return vfd;
}

void video_device_release(struct video_device *vfd)
{
	kfree(vfd);
}

#endif


/* initialitation*/
static struct video_device *vdev_init(int board, struct video_device *template, char *type)
{
	struct video_device *vfd;

	vfd = video_device_alloc();
	if (NULL == vfd)
		return NULL;
	*vfd = *template;

#if LINUX_VERSION_CODE < KERNEL_VERSION_V4L2DEV
#if LINUX_VERSION_CODE > KERNEL_VERSION(2, 5, 0)
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 27)
	vfd->dev     = &s616_dev[board]->dev;
#else
	vfd->parent     = &s616_dev[board]->dev;
#endif
#endif
#endif
	/*
	  dbgprint(vfd->name, sizeof(vfd->name), "BT%d%s %s (%s)",
	  btv->id, (btv->id==848 && btv->revision==0x12) ? "A" : "",
	  type, v16_tvcards[btv->c.type].name);
	*/
	return vfd;
}

/* register video for linux device*/
static int __devinit register_video(int board)
{
	int dev;
	struct sx16_v4l *v4l  = NULL;
	int rc;
	int type;
#if LINUX_VERSION_CODE > KERNEL_VERSION(5, 5, 0)
	type = VFL_TYPE_VIDEO;
#else
	type = VFL_TYPE_GRABBER;	
#endif
	vdev[board] = vdev_init(board, &v16_video_template, VIDEO_PREFIX);
	if (vdev[board] == NULL) {
		printk("%s failed\n", __func__);
		goto err;
	}
	v4l = kzalloc(sizeof(struct sx16_v4l), GFP_KERNEL);
	if (NULL == v4l) {
		printk(KERN_ERR "sx16: ENOMEM");
		return -ENOMEM;
	}
	video_set_drvdata(vdev[board], v4l);
#if LINUX_VERSION_CODE >= KERNEL_VERSION(4, 7, 0)	
	vdev[board]->device_caps = V4L2_CAP_DEVICE_CAPS | V4L2_CAP_VIDEO_CAPTURE;
#endif

#if LINUX_VERSION_CODE > KERNEL_VERSION_V4L2DEV
	strncpy(v4l->v4l2_dev.name, "sensoray 616 driver", sizeof(v4l->v4l2_dev.name));
	
	rc = v4l2_device_register(&s616_dev[board]->dev, &v4l->v4l2_dev);
	if (rc != 0) {
		goto err;
	}
	vdev[board]->v4l2_dev = &v4l->v4l2_dev;
#endif
	
	
	if (video_register_device(vdev[board], type, -1) < 0) {
		goto err;
	}


	
	dev = vdev[board]->minor & (MAX_DEVS - 1); /*0x1f;*/
	devnum_v4l[board] = dev;
	brdnum_v4l[dev] = board;
	printk(KERN_INFO "registered board %d as video%d\n", board, dev);
	G_v4l[board] = v4l;
	/* initialize initial v4l structures*/
	/* Initialize video4linux interface*/
	/* initialize structs / fill in defaults */
	v4l->pci = s616_dev[board];
				  
	spin_lock_init(&v4l->slock);
	mutex_init(&v4l->lock);
	INIT_LIST_HEAD(&v4l->video_q.queue);
#if LINUX_VERSION_CODE < KERNEL_VERSION(4, 4, 0)	
	init_timer(&v4l->video_q.timeout);
	v4l->video_q.timeout.function = sx16_buffer_timeout;
	v4l->video_q.timeout.data     = (unsigned long)(&v4l->video_q);
#elif LINUX_VERSION_CODE < KERNEL_VERSION(4, 8, 0)
	setup_timer(&v4l->video_q.timeout, sx16_buffer_timeout, (unsigned long) &v4l->video_q);
#else
	timer_setup(&v4l->video_q.timeout, sx16_buffer_timeout, 0);
#endif
	v4l->video_q.board = board;
	v4l->video_q.v4l = v4l;
	v4l->resources = 0;
#ifdef VIDIOC_G_PRIORITY
	v4l2_prio_init(&v4l->prio);
#endif
	v4l->board = board;
	return 0;
err:
	kfree(v4l);
	unregister_video(board);
	return -1;
}

/* unregister video for linux device*/
static void unregister_video(int board)
{
	if (vdev[board]) {
		if (vdev[board]->minor != -1)
			video_unregister_device(vdev[board]);
		else
			video_device_release(vdev[board]);
		vdev[board] = NULL;
	}
	v4l2_device_unregister(&G_v4l[board]->v4l2_dev);
	kfree(G_v4l[board]);
	G_v4l[board] = NULL;
}

#endif /* #if defined(SX16_V4L_BUILD)*/


#include <stdarg.h>

void dbgprint(char *fmt, ...)
{
#if LINUX_VERSION_CODE > KERNEL_VERSION(2, 6, 3)
	va_list argptr;
	static char str[256];

	va_start(argptr, fmt);
	vsprintf(str, fmt, argptr);
	va_end(argptr);
	printk(KERN_INFO "**%s\n", str);
#endif /*LINUX_VERSION_CODE > KERNEL_VERSION(2, 6, 3)*/
}

/* for 2.4 kernel and debug*/
void dbgprint3(char *fmt, int arg1, int arg2, int arg3)
{
	printk(fmt, arg1, arg2, arg3);
}

module_init(init_module_sx16);
module_exit(cleanup_module_sx16);
